import gc
import os
import json
import logging
from typing import Any, Dict, Optional, Text, List

import networkx as nx

from schematic.schemas.explorer import SchemaExplorer
from schematic.utils.io_utils import load_json
from schematic.utils.cli_utils import query_dict
from schematic.utils.schema_utils import load_schema_into_networkx
from schematic.utils.validate_utils import validate_schema, rule_in_rule_list

from schematic import CONFIG

logger = logging.getLogger(__name__)


class SchemaGenerator(object):
    def __init__(
        self,
        path_to_json_ld: str = None,
        schema_explorer: SchemaExplorer = None,
        requires_dependency_relationship: str = "requiresDependency",  # optional parameter(s) with default value
        requires_range: str = "rangeIncludes",
        range_value_relationship: str = "rangeValue",
        requires_component_relationship: str = "requiresComponent",
    ) -> None:
        """Create / Initialize object of type SchemaGenerator().

        Methods / utilities that are part of this module can be used to generate JSON validation schemas for different schema.org
        specification models.

        Args:
            path_to_json_ld: Path to the JSON-LD file that is representing the schema.org data model that we want to validate.
            schema_explorer: SchemaExplorer instance containing the schema.org data model that we want to validate.
            requires_dependency_relationship: Edge relationship between two nodes indicating that they are dependent on each other.
            requires_range: A node propertly indicating that a term can assume a value equal to any of the terms that are in the current term's range.
            range_value_relationship: Edge relationship that indicates a term / node that another node depends on, is part of the other node's range.
            requires_component_relationship: A node property indicating that this node requires a component for its full characterization.

        Returns:
            None
        """

        self.jsonld_path = path_to_json_ld

        if schema_explorer is None:

            assert (
                self.jsonld_path is not None
            ), "You must provide either `path_to_json_ld` or `schema_explorer`."

            self.jsonld_path_root, jsonld_ext = os.path.splitext(self.jsonld_path)

            assert jsonld_ext == ".jsonld", (
                "Please make sure the 'path_to_json_ld' parameter "
                "is pointing to a valid JSON-LD file."
            )

            # create an instance of SchemaExplorer
            self.se = SchemaExplorer()

            # convert the JSON-LD data model to networkx object
            self.se.load_schema(self.jsonld_path)

        else:

            # Confirm that given SchemaExplorer instance is valid
            assert (
                getattr(schema_explorer, "schema") is not None
                and getattr(schema_explorer, "schema_nx") is not None
            ), (
                "SchemaExplorer instance given to `schema_explorer` argument "
                "does not have both the `schema` and `schema_nx` attributes."
            )

            # User given instance of SchemaExplorer
            self.se = schema_explorer

        # custom value(s) of following relationship attributes are passed during initialization
        self.requires_dependency_relationship = requires_dependency_relationship
        self.requires_range = requires_range
        self.range_value_relationship = range_value_relationship
        self.requires_component_relationship = requires_component_relationship

    def get_edges_by_relationship(self, node: str, relationship: str) -> List[str]:
        """
        See class definition in SchemaExplorer
        TODO: possibly remove this wrapper and refactor downstream code to call from SchemaExplorer
        """

        return self.se.get_edges_by_relationship(node, relationship)

    def get_adjacent_nodes_by_relationship(
        self, node: str, relationship: str
    ) -> List[str]:

        """
        See class definition in SchemaExplorer
        TODO: possibly remove this wrapper and refactor downstream code to call from SchemaExplorer
        """

        return self.se.get_adjacent_nodes_by_relationship(node, relationship)

    def get_subgraph_by_edge_type(
        self, graph: nx.MultiDiGraph, relationship: str
    ) -> nx.DiGraph:
        """Get a subgraph containing all edges of a given type (aka relationship).
        TODO: possibly move method to SchemaExplorer and refactor downstream code to call from SchemaExplorer

        Args:
            graph: input multi digraph (aka hypergraph)
            relationship: edge / link relationship type with possible values same as in above docs.

        Returns:
            Directed graph on edges of a particular type (aka relationship)
        """

        # prune the metadata model graph so as to include only those edges that match the relationship type
        rel_edges = []
        for (u, v, key, c) in graph.out_edges(data=True, keys=True):
            if key == relationship:
                rel_edges.append((u, v))

        relationship_subgraph = nx.DiGraph()
        relationship_subgraph.add_edges_from(rel_edges)

        return relationship_subgraph

    def get_descendants_by_edge_type(
        self,
        source_node: str,
        relationship: str,
        connected: bool = True,
        ordered: bool = False,
    ) -> List[str]:

        """
        See class definition in SchemaExplorer
        TODO: possibly remove this wrapper and refactor downstream code to call from SchemaExplorer
        """

        return self.se.get_descendants_by_edge_type(
            source_node, relationship, connected, ordered
        )

    def get_component_requirements(self, source_component: str) -> List[str]:
        """Get all components that are associated with a given source component and are required by it.

        Args:
            source_component: source component for which we need to find all required downstream components.

        Returns:
            List of nodes that are descendants from the source component are are related to the source through a specific component relationship.
        """

        req_components = list(
            reversed(
                self.get_descendants_by_edge_type(
                    source_component, self.requires_component_relationship, ordered=True
                )
            )
        )

        return req_components

    def get_component_requirements_graph(self, source_component: str) -> nx.DiGraph:
        """Get all components that are associated with a given source component and are required by it; return the components as a dependency graph (i.e. a DAG).

        Args:
            source_component: source component for which we need to find all required downstream components.

        Returns:
            A subgraph of the schema graph induced on nodes that are descendants from the source component and are related to the source through a specific component relationship.
        """

        # get a list of required component nodes
        req_components = self.get_component_requirements(source_component)

        # get the schema graph
        mm_graph = self.se.get_nx_schema()

        # get the subgraph induced on required component nodes
        req_components_graph = self.get_subgraph_by_edge_type(
            mm_graph, self.requires_component_relationship
        ).subgraph(req_components)

        return req_components_graph

    def get_node_dependencies(
        self, source_node: str, display_names: bool = True, schema_ordered: bool = True
    ) -> List[str]:
        """Get the immediate dependencies that are related to a given source node.

        Args:
            source_node: The node whose dependencies we need to compute.
            display_names: if True, return list of display names of each of the dependencies.
                           if False, return list of node labels of each of the dependencies.
            schema_ordered: if True, return the dependencies of the node following the order of the schema (slower).
                            if False, return dependencies from graph without guaranteeing schema order (faster)

        Returns:
            List of nodes that are dependent on the source node.
        """
        mm_graph = self.se.get_nx_schema()

        if schema_ordered:
            # get dependencies in the same order in which they are defined in the schema
            required_dependencies = self.se.explore_class(source_node)["dependencies"]
        else:
            required_dependencies = self.get_adjacent_nodes_by_relationship(
                source_node, self.requires_dependency_relationship
            )

        if display_names:
            # get display names of dependencies
            dependencies_display_names = []

            for req in required_dependencies:
                dependencies_display_names.append(mm_graph.nodes[req]["displayName"])

            return dependencies_display_names

        return required_dependencies

    def get_node_range(self, node_label: str, display_names: bool = True) -> List[str]:
        """Get the range, i.e., all the valid values that are associated with a node label.

        Args:
            node_label: Node / termn for which you need to retrieve the range.

        Returns:
            List of display names of nodes associateed with the given node.
        """
        mm_graph = self.se.get_nx_schema()

        try:
            # get node range in the order defined in schema for given node
            required_range = self.se.explore_class(node_label)["range"]
        except KeyError:
            raise ValueError(
                f"The source node {node_label} does not exist in the graph. "
                "Please use a different node."
            )

        if display_names:
            # get the display name(s) of all dependencies
            dependencies_display_names = []

            for req in required_range:
                dependencies_display_names.append(mm_graph.nodes[req]["displayName"])

            return dependencies_display_names

        return required_range

    def get_node_label(self, node_display_name: str) -> str:
        """Get the node label for a given display name.

        Args:
            node_display_name: Display name of the node which you want to get the label for.

        Returns:
            Node label associated with given node.

        Raises:
            KeyError: If the node cannot be found in the graph.
        """
        mm_graph = self.se.get_nx_schema()

        node_class_label = self.se.get_class_label_from_display_name(node_display_name)
        node_property_label = self.se.get_property_label_from_display_name(
            node_display_name
        )

        if node_class_label in mm_graph.nodes:
            node_label = node_class_label
        elif node_property_label in mm_graph.nodes:
            node_label = node_property_label
        else:
            node_label = ""

        return node_label

    def get_node_definition(self, node_display_name: str) -> str:
        """Get the node definition, i.e., the "comment" associated with a given node display name.

        Args:
            node_display_name: Display name of the node which you want to get the label for.

        Returns:
            Comment associated with node, as a string.
        """
        node_label = self.get_node_label(node_display_name)

        if not node_label:
            return ""

        mm_graph = self.se.get_nx_schema()
        node_definition = mm_graph.nodes[node_label]["comment"]

        return node_definition

    def get_node_validation_rules(self, node_display_name: str) -> str:
        """Get validation rules associated with a node,

        Args:
            node_display_name: Display name of the node which you want to get the label for.

        Returns:
            A set of validation rules associated with node, as a list.
        """
        node_label = self.get_node_label(node_display_name)

        if not node_label:
            return []

        mm_graph = self.se.get_nx_schema()
        node_validation_rules = mm_graph.nodes[node_label]["validationRules"]

        return node_validation_rules

    def is_node_required(self, node_display_name: str) -> bool:
        """Check if a given node is required or not.

        Note: The possible options that a node can be associated with -- "required" / "optional".

        Args:
            node_display_name: Display name of the node which you want to get the label for.

        Returns:
            True: If the given node is a "required" node.
            False: If the given node is not a "required" (i.e., an "optional") node.
        """
        node_label = self.get_node_label(node_display_name)

        mm_graph = self.se.get_nx_schema()
        node_required = mm_graph.nodes[node_label]["required"]

        return node_required

    def get_nodes_display_names(
        self, node_list: List[str], mm_graph: nx.MultiDiGraph
    ) -> List[str]:
        """Get display names associated with the given list of nodes.

        Args:
            node_list: List of nodes whose display names we need to retrieve.

        Returns:
            List of display names.
        """
        node_list_display_names = [
            mm_graph.nodes[node]["displayName"] for node in node_list
        ]

        return node_list_display_names

    def get_range_schema(
        self, node_range: List[str], node_name: str, blank=False
    ) -> Dict[str, Dict[str, List[str]]]:
        """Add a list of nodes to the "enum" key in a given JSON schema object.

        Args:
            node_name: Name of the "main" / "head" key in the JSON schema / object.
            node_range: List of nodes to be added to the JSON object.
            blank: If True, add empty node to end of node list.
                   If False, do not add empty node to end of node list.

        Returns:
            JSON object with nodes.
        """
        if blank:
            schema_node_range = {node_name: {"enum": node_range + [""]}}
        else:
            schema_node_range = {node_name: {"enum": node_range}}

        return schema_node_range

    def get_array_schema(
        self, node_range: List[str], node_name: str, blank=False
    ) -> Dict[str, Dict[str, List[str]]]:
        """Add a list of nodes to the "enum" key in a given JSON schema object.
           Allow a node to be mapped to any subset of the list

        Args:
            node_name: Name of the "main" / "head" key in the JSON schema / object.
            node_range: List of nodes to be added to the JSON object.
            blank: If True, add empty node to end of node list.
                   If False, do not add empty node to end of node list.

        Returns:
            JSON object with array validation rule.
        """

        schema_node_range_array = {
            node_name: {
                "type": "array",
                "items": {"enum": node_range + [""] if blank else node_range},
                "maxItems": len(node_range),
            }
        }

        return schema_node_range_array

    def get_non_blank_schema(
        self, node_name: str
    ) -> Dict:  # can't define heterogenous Dict generic types
        """Get a schema rule that does not allow null or empty values.

        Args:
            node_name: Name of the node on which the schema rule is to be applied.

        Returns:
            Schema rule as a JSON object.
        """
        non_blank_schema = {node_name: {"not": {"type": "null"}, "minLength": 1}}

        return non_blank_schema

    def is_required(self, node_name: str, mm_graph: nx.MultiDiGraph) -> bool:
        """
        Check if a node is required

        Args:
            node_name: Name of the node on which the check is to be applied.

        Returns:
            Boolean value indicating if the node is required or not.
                True: yes, it is required.
                False: no, it is not required.
        """
        return mm_graph.nodes[node_name]["required"]

    def get_json_schema_requirements(self, source_node: str, schema_name: str) -> Dict:
        """Consolidated method that aims to gather dependencies and value constraints across terms / nodes in a schema.org schema and store them in a jsonschema /JSON Schema schema.

        It does so for any given node in the schema.org schema (recursively) using the given node as starting point in the following manner:
        1) Find all the nodes / terms this node depends on (which are required as "additional metadata" given this node is "required").
        2) Find all the allowable metadata values / nodes that can be assigned to a particular node (if such a constraint is specified on the schema).

        Args:
            source_node: Node from which we can start recursive dependancy traversal (as mentioned above).
            schema_name: Name assigned to JSON-LD schema (to uniquely identify it via URI when it is hosted on the Internet).

        Returns:
            JSON Schema as a dictionary.
        """
        json_schema = {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "$id": "http://example.com/" + schema_name,
            "title": schema_name,
            "type": "object",
            "properties": {},
            "required": [],
            "allOf": [],
        }

        # get graph corresponding to data model schema
        mm_graph = self.se.get_nx_schema()

        nodes_to_process = (
            []
        )  # list of nodes to be checked for dependencies, starting with the source node
        processed_nodes = (
            []
        )  # keep of track of nodes whose dependencies have been processed
        reverse_dependencies = (
            {}
        )  # maintain a map between conditional nodes and their dependencies (reversed) -- {dependency : conditional_node}
        range_domain_map = (
            {}
        )  # maintain a map between range nodes and their domain nodes {range_value : domain_value}
        # the domain node is very likely the parentof ("parentOf" relationship) of the range node

        root_dependencies = self.get_adjacent_nodes_by_relationship(
            source_node, self.requires_dependency_relationship
        )

        # if root_dependencies is empty it means that a class with name 'source_node' exists
        # in the schema, but it is not a valid component
        if not root_dependencies:
            raise ValueError(f"'{source_node}' is not a valid component in the schema.")

        nodes_to_process += root_dependencies

        process_node = nodes_to_process.pop(0)

        while process_node:

            if not process_node in processed_nodes:
                # node is being processed
                node_is_processed = True

                node_range = self.get_adjacent_nodes_by_relationship(
                    process_node, self.range_value_relationship
                )

                # get node range display name
                node_range_d = self.get_nodes_display_names(node_range, mm_graph)

                node_dependencies = self.get_adjacent_nodes_by_relationship(
                    process_node, self.requires_dependency_relationship
                )

                # get process node display name
                node_display_name = mm_graph.nodes[process_node]["displayName"]

                # updating map between node and node's valid values
                for n in node_range_d:
                    if not n in range_domain_map:
                        range_domain_map[n] = []
                    range_domain_map[n].append(node_display_name)

                # can this node be map to the empty set (if required no; if not required yes)
                # TODO: change "required" to different term, required may be a bit misleading (i.e. is the node required in the schema)
                node_required = self.is_required(process_node, mm_graph)

                # get any additional validation rules associated with this node (e.g. can this node be mapped to a list of other nodes)
                node_validation_rules = self.get_node_validation_rules(
                    node_display_name
                )

                if node_display_name in reverse_dependencies:
                    # if node has conditionals set schema properties and conditional dependencies
                    # set schema properties
                    if node_range:
                        # if process node has valid value range set it in schema properties
                        schema_valid_vals = self.get_range_schema(
                            node_range_d, node_display_name, blank=True
                        )

                        if node_validation_rules:
                            # if this node has extra validation rules process them
                            # TODO: abstract this into its own validation rule constructor/generator module/class
                            if rule_in_rule_list("list", node_validation_rules):
                                # if this node can be mapped to a list of nodes
                                # set its schema accordingly
                                schema_valid_vals = self.get_array_schema(
                                    node_range_d, node_display_name, blank=True
                                )

                    else:
                        # otherwise, by default allow any values
                        schema_valid_vals = {node_display_name: {}}

                    json_schema["properties"].update(schema_valid_vals)

                    # set schema conditional dependencies
                    for node in reverse_dependencies[node_display_name]:
                        # set all of the conditional nodes that require this process node

                        # get node domain if any
                        # ow this node is a conditional requirement
                        if node in range_domain_map:
                            domain_nodes = range_domain_map[node]
                            conditional_properties = {}

                            for domain_node in domain_nodes:

                                # set range of conditional node schema
                                conditional_properties.update(
                                    {
                                        "properties": {domain_node: {"enum": [node]}},
                                        "required": [domain_node],
                                    }
                                )

                                # given node conditional are satisfied, this process node (which is dependent on these conditionals) has to be set or not depending on whether it is required
                                if node_range:
                                    dependency_properties = self.get_range_schema(
                                        node_range_d,
                                        node_display_name,
                                        blank=not node_required,
                                    )

                                    if node_validation_rules:
                                        if rule_in_rule_list("list", node_validation_rules):
                                            # TODO: get_range_schema and get_range_schema have similar behavior - combine in one module
                                            dependency_properties = self.get_array_schema(
                                                node_range_d,
                                                node_display_name,
                                                blank=not node_required,
                                            )

                                else:
                                    if node_required:
                                        dependency_properties = self.get_non_blank_schema(
                                            node_display_name
                                        )
                                    else:
                                        dependency_properties = {node_display_name: {}}
                                schema_conditional_dependencies = {
                                    "if": conditional_properties,
                                    "then": {
                                        "properties": dependency_properties,
                                        "required": [node_display_name],
                                    },
                                }

                                # update conditional-dependency rules in json schema
                                json_schema["allOf"].append(
                                    schema_conditional_dependencies
                                )

                else:
                    # node doesn't have conditionals
                    if node_required:
                        if node_range:
                            schema_valid_vals = self.get_range_schema(
                                node_range_d, node_display_name, blank=False
                            )

                            if node_validation_rules:
                                # If there are valid values AND they are expected to be a list,
                                # reformat the Valid Values.
                                if rule_in_rule_list("list", node_validation_rules):
                                    schema_valid_vals = self.get_array_schema(
                                        node_range_d, node_display_name, blank=False
                                    )
                        else:
                            schema_valid_vals = self.get_non_blank_schema(
                                node_display_name
                            )

                        json_schema["properties"].update(schema_valid_vals)
                        # add node to required fields
                        json_schema["required"] += [node_display_name]

                    elif process_node in root_dependencies:
                        # node doesn't have conditionals and is not required; it belongs in the schema only if it is in root's dependencies

                        if node_range:
                            schema_valid_vals = self.get_range_schema(
                                node_range_d, node_display_name, blank=True
                            )

                            if node_validation_rules:
                                if rule_in_rule_list("list", node_validation_rules):
                                    schema_valid_vals = self.get_array_schema(
                                        node_range_d, node_display_name, blank=True
                                    )

                        else:
                            schema_valid_vals = {node_display_name: {}}

                        json_schema["properties"].update(schema_valid_vals)

                    else:
                        # node doesn't have conditionals and it is not required and it is not a root dependency
                        # the node doesn't belong in the schema
                        # do not add to processed nodes since its conditional may be traversed at a later iteration (though unlikely for most schemas we consider)
                        node_is_processed = False

                # add process node as a conditional to its dependencies
                node_dependencies_d = self.get_nodes_display_names(
                    node_dependencies, mm_graph
                )

                for dep in node_dependencies_d:
                    if not dep in reverse_dependencies:
                        reverse_dependencies[dep] = []

                    reverse_dependencies[dep].append(node_display_name)

                # add nodes found as dependencies and range of this processed node
                # to the list of nodes to be processed
                nodes_to_process += node_range
                nodes_to_process += node_dependencies

                # if the node is processed add it to the processed nodes set
                if node_is_processed:
                    processed_nodes.append(process_node)

            # if the list of nodes to process is not empty
            # set the process node the next remaining node to process
            if nodes_to_process:
                process_node = nodes_to_process.pop(0)
            else:
                # no more nodes to process
                # exit the loop
                break

        logger.info("JSON schema successfully generated from schema.org schema!")

        # if no conditional dependencies were added we can't have an empty 'AllOf' block in the schema, so remove it
        if not json_schema["allOf"]:
            del json_schema["allOf"]

        # Check if config value is provided; otherwise, set to None
        json_schema_log_file = query_dict(
            CONFIG.DATA, ("model", "input", "log_location")
        )

        # If no config value and SchemaGenerator was initialized with
        # a JSON-LD path, construct
        if json_schema_log_file is None and self.jsonld_path is not None:
            prefix = self.jsonld_path_root
            prefix_root, prefix_ext = os.path.splitext(prefix)
            if prefix_ext == ".model":
                prefix = prefix_root
            json_schema_log_file = f"{prefix}.{source_node}.schema.json"

        if json_schema_log_file is None:
            logger.info(
                "The JSON schema file can be inspected by setting the following "
                "nested key in the configuration: (model > input > log_location)."
            )
        else:
            json_schema_dirname = os.path.dirname(json_schema_log_file)
            if json_schema_dirname != '':
                os.makedirs(json_schema_dirname, exist_ok=True)
            with open(json_schema_log_file, "w") as js_f:
                json.dump(json_schema, js_f, indent=2)

        logger.info(f"JSON schema file log stored as {json_schema_log_file}")

        return json_schema



import os
import yaml


class Configuration(object):
    def __init__(self):
        # path to config.yml file
        self.CONFIG_PATH = None
        # entire configuration data
        self.DATA = None


    def __getattribute__(self, name):
        value = super().__getattribute__(name)
        if value is None and "SCHEMATIC_CONFIG_CONTENT" in os.environ:
            self.load_config_content_from_env()
            value = super().__getattribute__(name)
        elif value is None and "SCHEMATIC_CONFIG" in os.environ:
            self.load_config_from_env()
            value = super().__getattribute__(name)
        elif value is None and "SCHEMATIC_CONFIG" not in os.environ and "SCHEMATIC_CONFIG_CONTENT" not in os.environ:
            raise AttributeError(
                "The '%s' configuration field was accessed, but it hasn't been "
                "set yet, presumably because the schematic.CONFIG.load_config() "
                "method hasn't been run yet. Alternatively, you can re-run this "
                "code with the 'SCHEMATIC_CONFIG' environment variable set to "
                "the config.yml file, which will be automatically loaded." % name
            )
        return value

    def __getitem__(self, key):
        return self.DATA[key]

    def get(self, key, default):
        try:
            value = self[key]
        except AttributeError or KeyError:
            value = default
        return value

    def load_config_content(self, str_yaml: str) -> dict:
        try: 
            config_data = yaml.safe_load(str_yaml)
        except yaml.YAMLError as exc:
            print(exc)
            return None
        return config_data

    @staticmethod
    def load_yaml(file_path: str) -> dict:
        with open(file_path, "r") as stream:
            try:
                config_data = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                print(exc)
                return None
        return config_data

    def normalize_path(self, path):

        if self.CONFIG_PATH:
            # Retrieve parent directory of the config to decode relative paths
            parent_dir = os.path.dirname(self.CONFIG_PATH)
        else:
            # assume the parent dir would be the current work dir
            parent_dir = os.getcwd()

            # Ensure absolute file paths
        if not os.path.isabs(path):
            path = os.path.join(parent_dir, path)
        # And lastly, normalize file paths
        return os.path.normpath(path)

    def load_config_from_env(self):
        schematic_config = os.environ["SCHEMATIC_CONFIG"]
        print(
            "Loading config YAML file specified in 'SCHEMATIC_CONFIG' "
            "environment variable: %s" % schematic_config
        )
        return self.load_config(schematic_config)

    def load_config_content_from_env(self):
        schematic_config_content = os.environ["SCHEMATIC_CONFIG_CONTENT"]

        print(
            'Loading content of config file:  %s' % schematic_config_content
        )

        config_content_yaml = self.load_config_content(schematic_config_content)
        self.DATA = config_content_yaml

        return self.DATA

    def load_config(self, config_path=None, asset_view=None): 
        # If config_path is None, try loading from environment
        if config_path is None and "SCHEMATIC_CONFIG" in os.environ:
            return self.load_config_from_env()
        # Otherwise, raise an error
        elif config_path is None and "SCHEMATIC_CONFIG" not in os.environ:
            raise ValueError(
                "No configuration file provided to the `config_path` argument "
                "in `load_config`()`, nor was one specified in the "
                "'SCHEMATIC_CONFIG' environment variable. Quitting now..."
            )
        # Load configuration YAML file
        config_path = os.path.expanduser(config_path)
        config_path = os.path.abspath(config_path)
        self.DATA = self.load_yaml(config_path)
        self.CONFIG_PATH = config_path
        # handle user input (for API endpoints)
        if asset_view: 
            self.DATA['synapse']['master_fileview'] = asset_view

        # Return self.DATA as a side-effect
        return self.DATA
    @property
    def CREDS_PATH(self):
        self._CREDS_PATH = self.DATA["definitions"]["creds_path"]
        self._CREDS_PATH = self.normalize_path(self._CREDS_PATH)
        return self._CREDS_PATH

    @property
    def TOKEN_PICKLE(self):
        self._TOKEN_PICKLE = self.DATA["definitions"]["token_pickle"]
        self._TOKEN_PICKLE = self.normalize_path(self._TOKEN_PICKLE)
        return self._TOKEN_PICKLE

    @property
    def SERVICE_ACCT_CREDS(self):
        self._SERVICE_ACCT_CREDS = self.DATA["definitions"]["service_acct_creds"]
        self._SERVICE_ACCT_CREDS = self.normalize_path(self._SERVICE_ACCT_CREDS)
        return self._SERVICE_ACCT_CREDS

    @property
    def SYNAPSE_CONFIG_PATH(self):
        self._SYNAPSE_CONFIG_PATH = self.DATA["definitions"]["synapse_config"]
        self._SYNAPSE_CONFIG_PATH = self.normalize_path(self._SYNAPSE_CONFIG_PATH)
        return self._SYNAPSE_CONFIG_PATH


CONFIG = Configuration()



import os
import pickle
import logging

import pygsheets as ps

from typing import Dict, Any

from googleapiclient.discovery import build
from google_auth_oauthlib.flow import InstalledAppFlow
from google.auth.transport.requests import Request
from google.oauth2 import service_account
from google.oauth2.credentials import Credentials
from schematic import CONFIG
from schematic.store.synapse import SynapseStorage
import pandas as pd

logger = logging.getLogger(__name__)


# If modifying these scopes, delete the file token.pickle.
SCOPES = [
    "https://www.googleapis.com/auth/spreadsheets",
    "https://www.googleapis.com/auth/drive",
]


# it will create 'token.pickle' based on credentials.json
def generate_token() -> Credentials:
    creds = None
    # The file token.pickle stores the user's access and refresh tokens,
    # and is created automatically when the authorization flow completes for the first time.
    if os.path.exists(CONFIG.TOKEN_PICKLE):
        with open(CONFIG.TOKEN_PICKLE, "rb") as token:
            creds = pickle.load(token)

    # If there are no (valid) credentials available, let the user log in.
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            flow = InstalledAppFlow.from_client_secrets_file(CONFIG.CREDS_PATH, SCOPES)
            creds = flow.run_console()  ### don't have to deal with ports
        # Save the credentials for the next run
        with open(CONFIG.TOKEN_PICKLE, "wb") as token:
            pickle.dump(creds, token)

    return creds


# TODO: replace by pygsheets calls?
def build_credentials() -> Dict[str, Any]:
    creds = generate_token()

    # get a Google Sheet API service
    sheet_service = build("sheets", "v4", credentials=creds)
    # get a Google Drive API service
    drive_service = build("drive", "v3", credentials=creds)

    return {
        "sheet_service": sheet_service,
        "drive_service": drive_service,
        "creds": creds,
    }


def build_service_account_creds() -> Dict[str, Any]:
    credentials = service_account.Credentials.from_service_account_file(
        CONFIG.SERVICE_ACCT_CREDS, scopes=SCOPES
    )

    # get a Google Sheet API service
    sheet_service = build("sheets", "v4", credentials=credentials)
    # get a Google Drive API service
    drive_service = build("drive", "v3", credentials=credentials)

    return {
        "sheet_service": sheet_service,
        "drive_service": drive_service,
        "creds": credentials,
    }


def download_creds_file(auth: str = "token") -> None:
    if auth is None:
        raise ValueError(
            f"'{auth}' is not a valid authentication method. Please "
            "enter one of 'token' or 'service_account'."
        )

    syn = SynapseStorage.login()

    if auth == "token":
        if not os.path.exists(CONFIG.CREDS_PATH):
            # synapse ID of the 'credentials.json' file
            API_CREDS = CONFIG["synapse"]["token_creds"]

            # Download in parent directory of CREDS_PATH to
            # ensure same file system for os.rename()
            creds_dir = os.path.dirname(CONFIG.CREDS_PATH)

            creds_file = syn.get(API_CREDS, downloadLocation=creds_dir)
            os.rename(creds_file.path, CONFIG.CREDS_PATH)

            logger.info(
                "The credentials file has been downloaded " f"to '{CONFIG.CREDS_PATH}'"
            )

    elif auth == "service_account":
        if not os.path.exists(CONFIG.SERVICE_ACCT_CREDS):
            # synapse ID of the 'schematic_service_account_creds.json' file
            API_CREDS = CONFIG["synapse"]["service_acct_creds"]

            # Download in parent directory of SERVICE_ACCT_CREDS to
            # ensure same file system for os.rename()
            creds_dir = os.path.dirname(CONFIG.SERVICE_ACCT_CREDS)

            creds_file = syn.get(API_CREDS, downloadLocation=creds_dir)
            os.rename(creds_file.path, CONFIG.SERVICE_ACCT_CREDS)

            logger.info(
                "The credentials file has been downloaded "
                f"to '{CONFIG.SERVICE_ACCT_CREDS}'"
            )

    else:
        logger.warning(
            f"The mode of authentication you selected '{auth}' is "
            "not supported. Please use one of either 'token' or "
            "'service_account'."
        )


def execute_google_api_requests(service, requests_body, **kwargs):
    """
    Execute google API requests batch; attempt to execute in parallel.
    Args:
        service: google api service; for now assume google sheets service that is instantiated and authorized
        service_type: default batchUpdate; TODO: add logic for values update
        kwargs: google API service parameters
    Return: google API response
    """

    if (
        "spreadsheet_id" in kwargs
        and "service_type" in kwargs
        and kwargs["service_type"] == "batch_update"
    ):
        # execute all requests
        response = (
            service.spreadsheets()
            .batchUpdate(spreadsheetId=kwargs["spreadsheet_id"], body=requests_body)
            .execute()
        )

        return response

def export_manifest_drive_service(manifest_url, file_path, mimeType):
    '''
    Export manifest by using google drive api. If export as an Excel spreadsheet, the exported spreasheet would also include a hidden sheet 
    Args:
        manifest_url: google sheet manifest url
        file_path: file path of the exported manifest
        mimeType: exporting mimetype

    result: Google sheet gets exported in desired format

    '''

    # intialize drive service 
    services_creds = build_service_account_creds()
    drive_service = services_creds["drive_service"]

    # get spreadsheet id
    spreadsheet_id = manifest_url.split('/')[-1]

    # use google drive 
    data = drive_service.files().export(fileId=spreadsheet_id, mimeType=mimeType).execute()

    # open file and write data
    with open(os.path.abspath(file_path), 'wb') as f:
        try: 
            f.write(data)
        except FileNotFoundError as not_found: 
            logger.error(f"{not_found.filename} could not be found")

    f.close
   

def export_manifest_csv(file_path, manifest):
    '''
    Export manifest as a CSV by using google drive api
    Args:
        manifest: could be a dataframe or a manifest url
        file_path: file path of the exported manifest
        mimeType: exporting mimetype

    result: Google sheet gets exported as a CSV
    '''

    if isinstance(manifest, pd.DataFrame):
        manifest.to_csv(file_path, index=False)
    else: 
        export_manifest_drive_service(manifest, file_path, mimeType = 'text/csv')



def export_manifest_excel(manifest, output_excel=None):
    '''
    Export manifest as an Excel spreadsheet by using google sheet API. This approach could export hidden sheet
    Args:
        manifest: could be a dataframe or a manifest url 
        output_excel: name of the exported manifest sheet
    result: Google sheet gets exported as an excel spreadsheet. If there's a hidden sheet, the hidden sheet also gets exported. 
    '''
    # intialize drive service 
    services_creds = build_service_account_creds()
    sheet_service = services_creds["sheet_service"]

    if isinstance(manifest, pd.DataFrame):
        manifest.to_excel(output_excel, index=False)
    else:
        # get spreadsheet id from url 
        spreadsheet_id = manifest.split('/')[-1]

        # use google sheet api
        sheet_metadata = sheet_service.spreadsheets().get(spreadsheetId=spreadsheet_id).execute()
        sheets = sheet_metadata.get('sheets')

        # export to Excel
        writer = pd.ExcelWriter(output_excel)

        # export each sheet in manifest
        for sheet in sheets:
            dataset = sheet_service.spreadsheets().values().get(spreadsheetId=spreadsheet_id, range=sheet['properties']['title']).execute()
            dataset_df = pd.DataFrame(dataset['values'])
            dataset_df.columns = dataset_df.iloc[0]
            dataset_df.drop(dataset_df.index[0], inplace=True)
            dataset_df.to_excel(writer, sheet_name=sheet['properties']['title'], index=False)
        writer.save()
        writer.close()   


#!/usr/bin/env python3

# `schematic manifest` related sub-commands description
manifest_commands = {
    "manifest": {
        "config": (
            "Specify the path to the `config.yml` using this option. This is a required argument."
        ),
        "get": {
            "short_help": (
                "Specify the path to the `config.yml` using this option. "
                "This is a required argument."
            ),
            "title": (
                "Specify the title of the manifest (or title prefix of multiple manifests) that "
                "will be created at the end of the run. You can either explicitly pass the "
                "title of the manifest here or provide it in the `config.yml` "
                "file as a value for the `(manifest > title)` key."
            ),
            "data_type": (
                "Specify the component(s) (data type) from the data model that is to be used "
                "for generating the metadata manifest file. To make all available manifests enter 'all manifests'. "
                "You can either explicitly pass the data type here or provide "
                "it in the `config.yml` file as a value for the `(manifest > data_type)` key."
            ),
            "jsonld": (
                "Specify the path to the JSON-LD data model (schema) using this option. You can either explicitly pass the "
                "schema here or provide a value for the `(model > input > location)` key."
            ),
            "dataset_id": (
                "Specify the synID of a dataset folder on Synapse. If there is an exisiting manifest already present "
                "in that folder, then it will be pulled with the existing annotations for further annotation/modification. "
            ),
            "sheet_url": (
                "This is a boolean flag. If flag is provided when command line utility is executed, result will be a link/URL "
                "to the metadata manifest file. If not it will produce a pandas dataframe for the same."
            ),
            "output_csv": ("Path to where the CSV manifest template should be stored."),
            "output_xlsx": ("Path to where the Excel manifest template should be stored."),
            "use_annotations": (
                "This is a boolean flag. If flag is provided when command line utility is executed, it will prepopulate template "
                "with existing annotations from Synapse."
            ),
            "oauth": (
                "This is a boolean flag. If flag is provided when command line utility is executed, OAuth will be used to "
                "authenticate your Google credentials. If not service account mode of authentication will be used."
            ),
            "json_schema": (
                "Specify the path to the JSON Validation Schema for this argument. "
                "You can either explicitly pass the `.json` file here or provide it in the `config.yml` file "
                "as a value for the `(model > input > validation_schema)` key."
            ),
            "alphabetize_valid_values": (
                "Specify to alphabetize valid attribute values either ascending (a) or descending (d)." 
                "Optional"
            ),
        },
        "migrate": {
            "short_help": (
                "Specify the path to the `config.yml` using this option. "
                "This is a required argument."
            ),
            "project_scope": (
                "Specify a comma-separated list of projects where manifest entities will be migrated to tables."
            ),
            "archive_project": (
                "Specify a single project where legacy manifest entities will be stored after migration to table."
            ),
            "return_entities": (
                "This is a boolean flag. If flag is provided when command line utility is executed, "
                "entities that have been transferred to an archive project will be returned to their original folders."
            ),
            "dry_run": (
                "This is a boolean flag. If flag is provided when command line utility is executed, "
                "a dry run will be performed. No manifests will be re-uploaded and no entities will be migrated, "
                "but archival folders will still be created. "
                "Migration information for testing purposes will be logged to the INFO level."
            ),
        },
    }
}


# `schematic model` related sub-commands description
model_commands = {
    "model": {
        "config": (
            "Specify the path to the `config.yml` using this option. This is a required argument."
        ),
        "submit": {
            "short_help": ("Validation (optional) and submission of manifest files."),
            "manifest_path": (
                "Specify the path to the metadata manifest file that you want to submit to a dataset on Synapse. "
                "This is a required argument."
            ),
            "dataset_id": (
                "Specify the synID of the dataset folder on Synapse to which you intend to submit "
                "the metadata manifest file. This is a required argument."
            ),
            "validate_component": (
                "The component or data type from the data model which you can use to validate the "
                "data filled in your manifest template."
            ),
            "use_schema_label": (
                "Store attributes using the schema label (--use_schema_label, default) or store attributes using the display label "
                "(--use_display_label). Attribute display names in the schema must not only include characters that are "
                "not accepted by Synapse. Annotation names may only contain: letters, numbers, '_' and '.'"
            ),
            "hide_blanks":(
                "This is a boolean flag. If flag is provided when command line utility is executed, annotations with blank values will be hidden from a dataset's annotation list in Synaspe."
                "If not, annotations with blank values will be displayed."
            ),
            "manifest_record_type":(
                "Specify the way the manifest should be store as on Synapse. Options are 'entity', 'table' and "
                "'both'. 'entity' will store the manifest as a csv and create Synapse files for each row in the manifest. "
                "'table' will store the manifest as a table and a csv on Synapse. 'both' will do both of the options specified above. "
                "Default value is 'table'."
            ),      
        },
        "validate": {
            "short_help": ("Validation of manifest files."),
            "manifest_path": (
                "Specify the path to the metadata manifest file that you want to submit to a dataset on Synapse. "
                "This is a required argument."
            ),
            "data_type": (
                "Specify the component (data type) from the data model that is to be used "
                "for validating the metadata manifest file. You can either explicitly pass the data type here or provide "
                "it in the `config.yml` file as a value for the `(manifest > data_type)` key."
            ),
            "json_schema": (
                "Specify the path to the JSON Validation Schema for this argument. "
                "You can either explicitly pass the `.json` file here or provide it in the `config.yml` file "
                "as a value for the `(model > input > validation_schema)` key."
            ),
            "restrict_rules":(
                "This is a boolean flag. If flag is provided when command line utility is executed, validation suite will only run with in-house validation rules, "
                "and Great Expectations rules and suite will not be utilized."
                "If not, the Great Expectations suite will be utilized and all rules will be available."
            ),
            "project_scope": (
                "Specify a comma-separated list of projects to search through for cross manifest validation."
            ),
        },
    }
}


# `schematic schema` related sub-commands description
schema_commands = {
    "schema": {
        "convert": {
            "short_help": (
                "Convert specification from CSV data model to JSON-LD data model."
            ),
            "base_schema": (
                "Path to base data model. BioThings data model is loaded by default."
            ),
            "output_jsonld": (
                "Path to where the generated JSON-LD file needs to be outputted."
            ),
        }
    }
}


# `schematic init` command description
init_command = {
    "init": {
        "short_help": ("Initialize mode of authentication for schematic."),
        "auth": (
            "Specify the mode of authentication you want to use for Google accounts. "
            "You can use one of either 'token' or 'service_account'. The default mode of authentication "
            "is 'token' which uses OAuth."
        ),
        "config": (
            "Specify the path to the `config.yml` using this option. This is a required argument."
        ),
    }
}

viz_commands = {
    "visualization": {
        "config": (
            "Specify the path to the `config.yml` using this option. This is a required argument."
        ),
        "tangled_tree": {
            "figure_type": ("Specify the type of schema visualization to make. Either 'dependency' or 'component'."
                ),
            "text_format": ("Specify the type of text to gather for tangled tree visualization, either 'plain' or 'highlighted'."
                ),
        }
    }
    
}



#!/usr/bin/env python3

import inspect
import logging

from typing import Any, Mapping, Sequence, Union, List
from functools import reduce
import re

from schematic import CONFIG
from schematic.exceptions import (
    MissingConfigValueError,
    MissingConfigAndArgumentValueError,
)

logger = logging.getLogger(__name__)


def query_dict(dictionary: Mapping[Any, Any], keys: Sequence[Any]) -> Union[Any, None]:
    """Access a nested value in a dictionary corresponding
    to a series of keys.

    Args:
        dictionary: A dictionary containing anything.
        keys: A sequence of values corresponding to keys
            in `dictionary`

    Returns:
        The nested value corresponding to the given series
        of keys, or `None` is such a value doesn't exist.
    """

    def extract(dictionary: Any, key: Any) -> Union[Any, None]:
        """Get value associated with key, defaulting to None."""
        if dictionary is None or not isinstance(dictionary, dict):
            return None
        return dictionary.get(key)

    return reduce(extract, keys, dictionary)


def get_from_config(
    dictionary: Mapping[Any, Any], keys: Sequence[Any]
) -> Union[Any, None]:
    """Access a nested configuration value from a yaml
    configuration file.

    Args:
        dictionary: A dictionary containing anything.
        keys: A sequence of values corresponding to keys
            in `dictionary`.

    Returns:
        The nested value corresponding to the given series.

    Raises:
        MissingConfigValueError: When configuration value not
            found in config.yml file for given key.
    """
    # get configuration value from config file
    config_value = query_dict(dictionary, keys)

    # if configuration value not present then raise Exception
    if config_value is None:
        raise MissingConfigValueError(keys)

    config_keys_str = " > ".join(keys)

    logger.info(
        f"The ({config_keys_str}) argument with value "
        f"'{config_value}' is being read from the config file."
    )

    return config_value


def fill_in_from_config(
    arg_name: str, arg_value: Any, config_keys: Sequence[Any], allow_none: bool = False
) -> Any:
    """Fill in a missing value from a configuration object.

    Args:
        arg_name: Name of the argument. Used for logging.
        config_keys: List of keys used to access a nested
            value in `config` corresponding to `arg_name`.
        arg_value: Value of the argument provided at the
            command line.
        allow_none: Return None if argument value and
            configuration value are both None (rather
            than raising an error).

    Returns:
        The argument value, either from the calling context
        or the corresponding field in the configuration.

    Raises:
        AssertionError: If both the argument value and the
            configuration object are `None`.
    """

    # Avoid accessing config if argument value is provided
    if arg_value is not None:
        return arg_value

    # raise Exception if both, configuration value not present
    # in config file and CLI argument value is missing
    try:
        config_value = get_from_config(CONFIG.DATA, config_keys)
    except MissingConfigValueError:
        if allow_none:
            return None
        raise MissingConfigAndArgumentValueError(arg_name, config_keys)

    # Make sure argument value and
    config_keys_str = " > ".join(config_keys)

    logger.info(
        f"The '--{arg_name}' argument is being taken from configuration "
        f"file ({config_keys_str}), i.e., '{config_value}'."
    )

    return config_value

def parse_synIDs(
    ctx, param, synIDs,
) -> List[str]:
    """Parse and validate a comma separated string of synIDs

    Args:
        ctx:
            click option context
        param:
            click option argument name
        synIDs:
            comma separated string of synIDs

    Returns:
        List of synID strings

    Raises:
        ValueError: If the entire string does not match a regex for 
            a valid comma separated string of SynIDs
    """
    if synIDs:
        project_regex = re.compile("(syn\d+\,?)+")
        valid=project_regex.fullmatch(synIDs)

        if valid:
            synIDs = synIDs.split(",")

            return synIDs

        else:
            raise ValueError(
                        f"The provided list of project synID(s): {synIDs}, is not formatted correctly. "
                        "\nPlease check your list of projects for errors."
                    )
    else:
        return

def parse_comma_str_to_list(
    ctx, param, comma_string,
) -> List[str]:

    return comma_string.split(",")


#!/usr/bin/env python3

import logging
import sys

import click
import click_log

from schematic.visualization.attributes_explorer import AttributesExplorer
from schematic.visualization.tangled_tree import TangledTree
from schematic.utils.cli_utils import get_from_config, fill_in_from_config, query_dict
from schematic.help import viz_commands
from schematic.help import model_commands
from schematic import CONFIG

logger = logging.getLogger(__name__)
click_log.basic_config(logger)

CONTEXT_SETTINGS = dict(help_option_names=["--help", "-h"])  # help options

# invoke_without_command=True -> forces the application not to show aids before losing them with a --h
@click.group(context_settings=CONTEXT_SETTINGS, invoke_without_command=True)
@click_log.simple_verbosity_option(logger)
@click.option(
    "-c",
    "--config",
    type=click.Path(),
    envvar="SCHEMATIC_CONFIG",
    help=query_dict(model_commands, ("model", "config")),
)
@click.pass_context
def viz(ctx, config):  # use as `schematic model ...`
    """
    Sub-commands for Visualization methods.
    """
    try:
        logger.debug(f"Loading config file contents in '{config}'")
        ctx.obj = CONFIG.load_config(config)
    except ValueError as e:
        logger.error("'--config' not provided or environment variable not set.")
        logger.exception(e)
        sys.exit(1)

@viz.command(
    "attributes",
)
@click_log.simple_verbosity_option(logger)

@click.pass_obj
def get_attributes(ctx):
    """
    
    """
    # Get JSONLD file path
    path_to_jsonld = get_from_config(CONFIG.DATA, ("model", "input", "location"))
    # Run attributes explorer
    AttributesExplorer(path_to_jsonld).parse_attributes(save_file=True)
    return

@viz.command(
    "tangled_tree_text"
)
@click_log.simple_verbosity_option(logger)
@click.option(
    "-ft",
    "--figure_type",
    type=click.Choice(['component', 'dependency'], case_sensitive=False),
    help=query_dict(viz_commands, ("visualization", "tangled_tree", "figure_type")),
)
@click.option(
    "-tf",
    "--text_format",
    type=click.Choice(['plain', 'highlighted'], case_sensitive=False),
    help=query_dict(viz_commands, ("visualization", "tangled_tree", "text_format")),
)

@click.pass_obj
def get_tangled_tree_text(ctx, figure_type, text_format):
    """ Get text to be placed on the tangled tree visualization.
    """
    # Get JSONLD file path
    path_to_jsonld = get_from_config(CONFIG.DATA, ("model", "input", "location"))
    
    # Initialize TangledTree
    tangled_tree = TangledTree(path_to_jsonld, figure_type)

    # Get text for tangled tree.
    text_df = tangled_tree.get_text_for_tangled_tree(text_format, save_file=True)
    return

@viz.command(
    "tangled_tree_layers"
)
@click_log.simple_verbosity_option(logger)
@click.option(
    "-ft",
    "--figure_type",
    type=click.Choice(['component', 'dependency'], case_sensitive=False),
    help=query_dict(viz_commands, ("visualization", "tangled_tree", "figure_type")),
)

@click.pass_obj
def get_tangled_tree_component_layers(ctx, figure_type):
    ''' Get the components that belong in each layer of the tangled tree visualization.
    '''
    # Get JSONLD file path
    path_to_jsonld = get_from_config(CONFIG.DATA, ("model", "input", "location"))
    
    # Initialize Tangled Tree
    tangled_tree = TangledTree(path_to_jsonld, figure_type)
    
    # Get tangled trees layers JSON.
    layers = tangled_tree.get_tangled_tree_layers(save_file=True)

    return



#!/usr/bin/env python3

import click
import click_log
import logging
import sys
import re

from schematic.schemas.df_parser import _convert_csv_to_data_model
from schematic.utils.cli_utils import query_dict
from schematic.help import schema_commands

logger = logging.getLogger(__name__)
click_log.basic_config(logger)

CONTEXT_SETTINGS = dict(help_option_names=["--help", "-h"])  # help options

# invoke_without_command=True -> forces the application not to show aids before losing them with a --h
@click.group(context_settings=CONTEXT_SETTINGS, invoke_without_command=True)
def schema():  # use as `schematic model ...`
    """
    Sub-commands for Schema related utilities/methods.
    """
    pass


# prototype based on submit_metadata_manifest()
@schema.command(
    "convert",
    options_metavar="<options>",
    short_help=query_dict(schema_commands, ("schema", "convert", "short_help")),
)
@click_log.simple_verbosity_option(logger)
@click.argument(
    "schema_csv", type=click.Path(exists=True), metavar="<DATA_MODEL_CSV>", nargs=1
)
@click.option(
    "--base_schema",
    "-b",
    type=click.Path(exists=True),
    metavar="<JSON-LD_SCHEMA>",
    help=query_dict(schema_commands, ("schema", "convert", "base_schema")),
)
@click.option(
    "--output_jsonld",
    "-o",
    metavar="<OUTPUT_PATH>",
    help=query_dict(schema_commands, ("schema", "convert", "output_jsonld")),
)
def convert(schema_csv, base_schema, output_jsonld):
    """
    Running CLI to convert data model specification in CSV format to
    data model in JSON-LD format.
    """
    # convert RFC to Data Model
    base_se = _convert_csv_to_data_model(schema_csv, base_schema)

    # output JSON-LD file alongside CSV file by default
    if output_jsonld is None:
        csv_no_ext = re.sub("[.]csv$", "", schema_csv)
        output_jsonld = csv_no_ext + ".jsonld"

        logger.info(
            "By default, the JSON-LD output will be stored alongside the first "
            f"input CSV file. In this case, it will appear here: '{output_jsonld}'. "
            "You can use the `--output_jsonld` argument to specify another file path."
        )

    # saving updated schema.org schema
    try:
        base_se.export_schema(output_jsonld)
        click.echo(f"The Data Model was created and saved to '{output_jsonld}' location.")
    except:
        click.echo(f"The Data Model could not be created by using '{output_jsonld}' location. Please check your file path again")




from ast import arg
from jsonschema import ValidationError
import logging
import pandas as pd
from typing import Any, Dict, Optional, Text, List


logger = logging.getLogger(__name__)

def validation_rule_info():
    '''
    Function to return dict that holds information about each rule
    Will be pulled into validate_single_rule, validate_manifest_rules, validate_schema_rules
    Structure:    
        Rule:{
            'arguments':(<num arguments allowed>, <num arguments required>),
            'type': <rule type>,
            'complementary_rules': [<rules available for pairing>]}
        }
    '''
    rule_dict = {
            "int": {
                'arguments':(1, 0),
                'type': "type_validation",
                'complementary_rules': ['inRange',],
                'default_message_level': 'error'},

            "float": {
                'arguments':(1, 0),
                'type': "type_validation",
                'complementary_rules': ['inRange',],
                'default_message_level': 'error'},

            "num": {
                'arguments':(1, 0),
                'type': "type_validation",
                'complementary_rules': ['inRange',],
                'default_message_level': 'error'},

            "str": {
                'arguments':(1, 0),
                'type': "type_validation",
                'complementary_rules': None,
                'default_message_level': 'error'},

            "regex": {
                'arguments':(3, 2), 
                'fixed_arg': ['strict'], 
                'type': "regex_validation",
                'complementary_rules': ['list'],
                'default_message_level': 'error'},

            "url" : {
                'arguments':(101, 0), 
                'type': "url_validation",
                'complementary_rules': None,
                'default_message_level': 'error'},

            "list": {
                'arguments':(2, 0), 
                'type': "list_validation",
                'complementary_rules': ['regex'],
                'default_message_level': 'error'},
                
            "matchAtLeastOne": {
                'arguments':(3, 2), 
                'type': "cross_validation",
                'complementary_rules': None,
                'default_message_level': 'warning'},

            "matchExactlyOne": {
                'arguments':(3, 2), 
                'type': "cross_validation",
                'complementary_rules': None,
                'default_message_level': 'warning'},
                
            "recommended": {
                'arguments':(1, 0), 
                'type': "content_validation",
                'complementary_rules': None,
                'default_message_level': 'warning'},

            "protectAges": {
                'arguments':(1, 0), 
                'type': "content_validation",
                'complementary_rules': ['inRange',],
                'default_message_level': 'warning'},

            "unique": {
                'arguments':(1, 0), 
                'type': "content_validation",
                'complementary_rules': None,
                'default_message_level': 'error'},
                
            "inRange": {
                'arguments':(3, 2), 
                'type': "content_validation",
                'complementary_rules': ['int','float','num','protectAges'],
                'default_message_level': 'error'},
            }

    return rule_dict

def get_error(validation_rules: list, 
        attribute_name: str, error_type: str, input_filetype:str,) -> List[str]:
    '''
    Generate error message for errors when trying to specify 
    multiple validation rules.
    '''
    error_col = attribute_name # Attribute name
    if error_type == 'too_many_rules':
        error_str = (f"The {input_filetype}, has an error in the validation rule "
            f"for the attribute: {attribute_name}, the provided validation rules ({validation_rules}) ."
        f"have too many entries. We currently only allow for pairs of rules to be used at the same time.")
        logging.error(error_str)
        error_message = error_str
        error_val = f"Multiple Rules: too many rules"
    
    if error_type == 'delimiter':
        error_str = (f"The {input_filetype}, has an error in the validation rule "
            f"for the attribute: {attribute_name}, the provided validation rules ({validation_rules}) are improperly "
            f"specified. Please check your delimiter is '::'")
        logging.error(error_str)
        error_message = error_str
        error_val = f"Multiple Rules: Delimiter"
    
    if error_type == 'not_rule':
        error_str = (f"The {input_filetype}, has an error in the validation rule "
            f"for the attribute: {attribute_name}, the provided validation rules ({validation_rules}) is not "
            f"a valid rule. Please check spelling.")
        logging.error(error_str)
        error_message = error_str
        error_val = f"Not a Rule"
    
    if error_type == 'args_not_allowed':
        error_str = (f"The {input_filetype}, has an error in the validation rule "
            f"for the attribute: {attribute_name}, the provided validation rules ({validation_rules}) is not"
            f"formatted properly. No additional arguments are allowed for this rule.")
        logging.error(error_str)
        error_message = error_str
        error_val = f"Args not allowed."
    if error_type == 'incorrect_num_args':
        rule_type=validation_rules.split(" ")[0]
        
        if rule_type in validation_rule_info():
            no_allowed, no_required = validation_rule_info()[rule_type]['arguments']
        else:
            no_allowed, no_required = ('', '')

        error_str = (f"The {input_filetype}, has an error in the validation rule "
            f"for the attribute: {attribute_name}, the provided validation rules ({validation_rules}) is not "
            f"formatted properly. The number of provided arguments does not match the number allowed({no_allowed}) or required({no_required}).")
        logging.error(error_str)
        error_message = error_str
        error_val = f"Incorrect num arguments."

    if error_type == 'invalid_rule_combination':
        first_type=validation_rules[0].split(' ')[0]
        second_type=validation_rule_info()[first_type]['complementary_rules']

        error_str = (f"The {input_filetype}, has an error in the validation rule "
            f"for the attribute: {attribute_name}, the provided validation rules ({'::'.join(validation_rules)}) are not "
            f"a valid combination of rules. The validation rule [{first_type}] may only be used with rules of type {second_type}.")
        logging.error(error_str)
        error_message = error_str
        error_val = f"Invalid rule combination."
        
    return ['NA', error_col, error_message, error_val]

def validate_single_rule(validation_rule, attribute, input_filetype):
    '''
    TODO:reformat validation_types from validate_manifest to document the 
    arguments allowed. Keep in that location and pull into this function.
    '''
    errors = []
    validation_types = validation_rule_info()
    validation_rule_with_args = [
                val_rule.strip() for val_rule in validation_rule.strip().split(" ")]

    rule_type = validation_rule_with_args[0]

    # Check that the rule is actually a valid rule type.
    if rule_type not in validation_types.keys():
        errors.append(get_error(validation_rule, attribute,
            error_type = 'not_rule', input_filetype=input_filetype))

    else:
        arguments_allowed, arguments_required = validation_types[rule_type]['arguments']
        # Remove any fixed args from our calc.
        if 'fixed_arg' in validation_types[rule_type].keys():
            fixed_args = validation_types[rule_type]['fixed_arg']
            num_args = len([vr for vr in validation_rule_with_args if vr not in fixed_args])-1 
        else:
            num_args = len(validation_rule_with_args) - 1
            
        # If arguments are provided but not allowed, raise an error.
        if num_args and not arguments_allowed:
            errors.append(get_error(validation_rule, attribute,
                error_type = 'args_not_allowed', input_filetype=input_filetype))
        
        # If arguments are allowed, check that the correct amount have been passed.
        # There must be at least the number of args required,
        # and not more than allowed
        elif arguments_allowed:
            if (num_args < arguments_required) or (num_args > arguments_allowed):
                errors.append(get_error(validation_rule, attribute,
                    error_type = 'incorrect_num_args', input_filetype=input_filetype))


        if ':' in validation_rule:
            errors.append(get_error(validation_rule, attribute,
                error_type = 'delimiter', input_filetype=input_filetype))
    return errors

def validate_schema_rules(validation_rules, attribute, input_filetype):
    '''
    validation_rules: list
    input_filetype: str, used in error generation to aid user in
        locating the source of the error.

    Validation Rules Formatting rules:
    Multiple Rules:
        max of 2 rules
    Single Rules:
        Additional arg
    '''
    # Specify all the validation types and whether they currently
    # allow users to pass additional arguments (when used on their own), and if there is
    # a set number of arguments required.

    errors = []
    rule_info = validation_rule_info()
    num_validation_rules = len(validation_rules)

    
    # Check for edge case that user has entered more than 2 rules,
    # throw an error if they have.
    if num_validation_rules > 2:
            errors.append(get_error(validation_rules, attribute,
                error_type = 'too_many_rules', input_filetype=input_filetype))

    elif num_validation_rules == 2: 
        first_rule, second_rule = validation_rules
        first_type = first_rule.split(" ")[0]  
        second_type = second_rule.split(" ")[0]  

        # validate rule combination
        if second_type not in rule_info[first_type]['complementary_rules']:
            errors.append(get_error(validation_rules, attribute, 
                error_type = 'invalid_rule_combination', input_filetype=input_filetype))
        
    # validate each rule individually in the combo
    for rule in validation_rules:
        errors.extend(validate_single_rule(rule,
            attribute, input_filetype))
                

    if errors:
        raise ValidationError(
                        f"The {input_filetype} has an error in the validation_rules set "
                        f"for attribute {attribute}. "
                        f"Validation failed with the following errors: {errors}"
                    )
    
    return 


from schematic.utils.curie_utils import (
    expand_curie_to_uri,
    expand_curies_in_schema,
    extract_name_from_uri_or_curie,
    uri2label,
)
from schematic.utils.df_utils import update_df
from schematic.utils.general import dict2list, find_duplicates, str2list, unlist
from schematic.utils.google_api_utils import (
    download_creds_file,
    execute_google_api_requests,
    export_manifest_csv,
    export_manifest_excel,
)
from schematic.utils.io_utils import (
    export_json,
    load_default,
    load_json,
    load_schemaorg,
)
from schematic.utils.schema_utils import load_schema_into_networkx
from schematic.utils.validate_utils import (
    validate_class_schema,
    validate_property_schema,
    validate_schema,
)
from schematic.utils.viz_utils import visualize



# allows specifying explicit variable types
from typing import Any, Dict, Optional, Text


def find_duplicates(_list):
    """Find duplicate items in a list"""
    return set([x for x in _list if _list.count(x) > 1])


def dict2list(dictionary):
    if type(dictionary) == list:
        return dictionary
    elif type(dictionary) == dict:
        return [dictionary]


def str2list(_str):
    if type(_str) == str:
        return [_str]
    elif type(_str) == list:
        return _str


def unlist(_list):
    if len(_list) == 1:
        return _list[0]
    else:
        return _list



#!/usr/bin/env python

import logging
import click
import click_log

from schematic.manifest.commands import (
    manifest as manifest_cli,
)  # get manifest commands
from schematic.models.commands import model as model_cli  # submit manifest commands
from schematic.schemas.commands import (
    schema as schema_cli,
)  # schema conversion commands
from schematic.visualization.commands import (
    viz as viz_cli,
)  # viz generation commands
from schematic import init as init_cli  # schematic initialization commands

logger = logging.getLogger()
click_log.basic_config(logger)

# dict() -> new empty dictionary
CONTEXT_SETTINGS = dict(help_option_names=["--help", "-h"])  # help options

# invoke_without_command=True -> forces the application not to show aids before losing them with a --h
@click.group(context_settings=CONTEXT_SETTINGS, invoke_without_command=True)
@click_log.simple_verbosity_option(logger)
def main():
    """
    Command line interface to the `schematic` backend services.
    """
    logger.info("Starting schematic...")
    logger.debug("Existing sub-commands need to be used with schematic.")


main.add_command(init_cli)  # add init commands
main.add_command(manifest_cli)  # add manifest commands
main.add_command(model_cli)  # add model commands
main.add_command(schema_cli)  # add schema commands
main.add_command(viz_cli)  # add viz commands



if __name__ == "__main__":
    main()



import networkx as nx
import json

from schematic.utils.curie_utils import extract_name_from_uri_or_curie
from schematic.utils.validate_utils import validate_class_schema
from schematic.utils.validate_rules_utils import validate_schema_rules


def load_schema_into_networkx(schema):
    G = nx.MultiDiGraph()
    for record in schema["@graph"]:

        # TODO: clean up obsolete code
        # if record["@type"] == "rdfs:Class":

        # creation of nodes
        # adding nodes to the graph
        node = {}
        for (k, value) in record.items():
            # Some keys in the current schema.org schema have a dictionary entry for their value that includes keys @language and @value, 
            # for parity with other schemas, we just want the value
            if isinstance(value,dict) and "@language" in value.keys():
                record[k] = record[k]["@value"]
            if ":" in k:
                key = k.split(":")[1]
                node[key] = value
            elif "@" in k:
                key = k[1:]
                node[key] = value
            else:
                node[k] = value

        # creation of edges
        # adding edges to the graph
        if "rdfs:subClassOf" in record:
            parents = record["rdfs:subClassOf"]
            if type(parents) == list:
                for _parent in parents:
                    n1 = extract_name_from_uri_or_curie(_parent["@id"])
                    n2 = record["rdfs:label"]

                    # do not allow self-loops
                    if n1 != n2:
                        G.add_edge(n1, n2, key="parentOf")
            elif type(parents) == dict:
                n1 = extract_name_from_uri_or_curie(parents["@id"])
                n2 = record["rdfs:label"]

                # do not allow self-loops
                if n1 != n2:
                    G.add_edge(n1, n2, key="parentOf")

        # TODO: refactor: abstract adding relationship method
        if "sms:requiresDependency" in record:
            dependencies = record["sms:requiresDependency"]
            if type(dependencies) == list:
                for _dep in dependencies:
                    n1 = record["rdfs:label"]
                    n2 = extract_name_from_uri_or_curie(_dep["@id"])
                    # do not allow self-loops
                    if n1 != n2:
                        G.add_edge(n1, n2, key="requiresDependency")

        if "sms:requiresComponent" in record:
            components = record["sms:requiresComponent"]
            if type(components) == list:
                for _comp in components:
                    n1 = record["rdfs:label"]
                    n2 = extract_name_from_uri_or_curie(_comp["@id"])
                    # do not allow self-loops
                    if n1 != n2:
                        G.add_edge(n1, n2, key="requiresComponent")

        if "schema:rangeIncludes" in record:
            range_nodes = record["schema:rangeIncludes"]
            if type(range_nodes) == list:
                for _range_node in range_nodes:
                    n1 = record["rdfs:label"]
                    n2 = extract_name_from_uri_or_curie(_range_node["@id"])
                    # do not allow self-loops
                    if n1 != n2:
                        G.add_edge(n1, n2, key="rangeValue")
            elif type(range_nodes) == dict:
                n1 = record["rdfs:label"]
                n2 = extract_name_from_uri_or_curie(range_nodes["@id"])
                # do not allow self-loops
                if n1 != n2:
                    G.add_edge(n1, n2, key="rangeValue")

        if "schema:domainIncludes" in record:
            domain_nodes = record["schema:domainIncludes"]
            if type(domain_nodes) == list:
                for _domain_node in domain_nodes:
                    n1 = extract_name_from_uri_or_curie(_domain_node["@id"])
                    n2 = record["rdfs:label"]
                    # do not allow self-loops
                    if n1 != n2:
                        G.add_edge(n1, n2, key="domainValue")
            elif type(domain_nodes) == dict:
                n1 = extract_name_from_uri_or_curie(domain_nodes["@id"])
                n2 = record["rdfs:label"]
                # do not allow self-loops
                if n1 != n2:
                    G.add_edge(n1, n2, key="domainValue")

        # check schema generator (JSON validation schema gen)
        if (
            "requiresChildAsValue" in node
            and node["requiresChildAsValue"]["@id"] == "sms:True"
        ):
            node["requiresChildAsValue"] = True

        if "required" in node:
            if "sms:true" == record["sms:required"]:
                node["required"] = True
            else:
                node["required"] = False

        # not sure if this is required?
        if "sms:validationRules" in record:
            node["validationRules"] = record["sms:validationRules"]
            if node["validationRules"]:
                validate_vr = validate_schema_rules(
                                record["sms:validationRules"],
                                record["rdfs:label"],
                                input_filetype = 'json_schema')
        else:
            node["validationRules"] = []

        node["uri"] = record["@id"]
        node["description"] = record["rdfs:comment"]
        G.add_node(record["rdfs:label"], **node)
        # print(node)
        # print(G.nodes())

    return G


def node_attrs_cleanup(class_add_mod: dict) -> dict:
    # clean map that will be inputted into the node/graph
    node = {}
    for (k, value) in class_add_mod.items():
        if ":" in k:
            key = k.split(":")[1]
            node[key] = value
        elif "@" in k:
            key = k[1:]
            node[key] = value
        else:
            node[k] = value

    return node


def relationship_edges(
    schema_graph_nx: nx.MultiDiGraph, class_add_mod: dict, **kwargs
) -> nx.MultiDiGraph:
    """
    Notes:
    =====
    # pass the below dictionary as the third argument (kwargs) to relationship_edges().
    # "in" indicates that the relationship has an in-edges behaviour.
    # "out" indicates that the relationship has an out-edges behaviour.

    rel_dict = {
        "rdfs:subClassOf": {
            "parentOf": "in"
        },
        "schema:domainIncludes": {
            "domainValue": "in"
        },
        "sms:requiresDependency": {
            "requiresDependency": "out"
        },
        "sms:requiresComponent": {
            "requiresComponent": "out"
        },
        "schema:rangeIncludes": {
            "rangeValue": "out"
        }
    }
    """
    for rel, rel_lab_node_type in kwargs.items():
        for rel_label, node_type in rel_lab_node_type.items():
            if rel in class_add_mod:
                parents = class_add_mod[rel]
                if type(parents) == list:
                    for _parent in parents:

                        if node_type == "in":
                            n1 = extract_name_from_uri_or_curie(_parent["@id"])
                            n2 = class_add_mod["rdfs:label"]

                        if node_type == "out":
                            n1 = class_add_mod["rdfs:label"]
                            n2 = extract_name_from_uri_or_curie(_parent["@id"])

                        # do not allow self-loops
                        if n1 != n2:
                            schema_graph_nx.add_edge(n1, n2, key=rel_label)
                elif type(parents) == dict:
                    if node_type == "in":
                        n1 = extract_name_from_uri_or_curie(parents["@id"])
                        n2 = class_add_mod["rdfs:label"]

                    if node_type == "out":
                        n1 = class_add_mod["rdfs:label"]
                        n2 = extract_name_from_uri_or_curie(parents["@id"])

                    # do not allow self-loops
                    if n1 != n2:
                        schema_graph_nx.add_edge(n1, n2, key=rel_label)

    return schema_graph_nx


def class_to_node(class_to_convert: dict) -> nx.Graph:
    G = nx.Graph()

    node = {}  # node to be added the above graph and returned
    for (k, v) in class_to_convert.items():
        if ":" in k:  # if ":" is present in key
            key = k.split(":")[1]
            node[key] = v
        elif "@" in k:  # if "@" is present in key
            key = k[1:]
            node[key] = v
        else:
            node[k] = v

    if "required" in node:
        if class_to_convert["sms:required"] == "sms:true":
            node["required"] = True
        else:
            node["required"] = False

    if "sms:validationRules" in class_to_convert:
        node["validationRules"] = class_to_convert["sms:validationRules"]
    else:
        node["validationRules"] = []

    node["uri"] = class_to_convert["@id"]  # add separate "uri" key
    node["description"] = class_to_convert[
        "rdfs:comment"
    ]  # separately store "comment" as "description"
    G.add_node(class_to_convert["rdfs:label"], **node)

    return G


def replace_node_in_schema(schema: nx.MultiDiGraph, class_add_mod: dict) -> None:
    # part of the code that replaces the modified class in the original JSON-LD schema (not in the data/ folder though)
    for i, schema_class in enumerate(schema["@graph"]):
        if schema_class["rdfs:label"] == class_add_mod["rdfs:label"]:
            validate_class_schema(
                class_add_mod
            )  # validate that the class to be modified follows the structure for any generic class (node)

            schema["@graph"][i] = class_add_mod
            break


def export_schema(schema, file_path):
    with open(file_path, "w") as f:
        json.dump(schema, f, sort_keys=True, indent=4, ensure_ascii=False)



import graphviz


def visualize(edges, size=None):
    if size:
        d = graphviz.Digraph(graph_attr=[("size", size)])
    else:
        d = graphviz.Digraph()

    for _item in edges:
        d.edge(_item[0], _item[1])
    return d



import os
import logging
from pathlib import Path
import click
import click_log
import logging
import sys
from typing import List

from schematic.manifest.generator import ManifestGenerator
from schematic.utils.cli_utils import fill_in_from_config, query_dict, parse_synIDs
from schematic.help import manifest_commands
from schematic import CONFIG
from schematic.schemas.generator import SchemaGenerator
from schematic.utils.google_api_utils import export_manifest_csv, export_manifest_excel, export_manifest_drive_service
from schematic.store.synapse import SynapseStorage

logger = logging.getLogger(__name__)
click_log.basic_config(logger)

CONTEXT_SETTINGS = dict(help_option_names=["--help", "-h"])  # help options


# invoke_without_command=True -> forces the application not to show aids before losing them with a --h
@click.group(context_settings=CONTEXT_SETTINGS, invoke_without_command=True)
@click_log.simple_verbosity_option(logger)
@click.option(
    "-c",
    "--config",
    envvar="SCHEMATIC_CONFIG",
    help=query_dict(manifest_commands, ("manifest", "config")),
)
@click.pass_context
def manifest(ctx, config):  # use as `schematic manifest ...`
    """
    Sub-commands with Manifest Generation utilities/methods.
    """
    try:
        logger.debug(f"Loading config file contents in '{config}'")
        ctx.obj = CONFIG.load_config(config)
    except ValueError as e:
        logger.error("'--config' not provided or environment variable not set.")
        logger.exception(e)
        sys.exit(1)


# prototype based on getModelManifest() and get_manifest()
# use as `schematic config get positional_args --optional_args`
@manifest.command(
    "get", short_help=query_dict(manifest_commands, ("manifest", "get", "short_help"))
)
@click_log.simple_verbosity_option(logger)
# define the optional arguments
@click.option(
    "-t", "--title", help=query_dict(manifest_commands, ("manifest", "get", "title"))
)
@click.option(
    "-dt",
    "--data_type",
    help=query_dict(manifest_commands, ("manifest", "get", "data_type")),
)
@click.option(
    "-p", "--jsonld", help=query_dict(manifest_commands, ("manifest", "get", "jsonld"))
)
@click.option(
    "-d",
    "--dataset_id",
    help=query_dict(manifest_commands, ("manifest", "get", "dataset_id")),
)
@click.option(
    "-s",
    "--sheet_url",
    is_flag=True,
    help=query_dict(manifest_commands, ("manifest", "get", "sheet_url")),
)
@click.option(
    "-o",
    "--output_csv",
    help=query_dict(manifest_commands, ("manifest", "get", "output_csv")),
)
@click.option(
    "-oxlsx",
    "--output_xlsx",
    help=query_dict(manifest_commands, ("manifest", "get", "output_xlsx")),
)
@click.option(
    "-a",
    "--use_annotations",
    is_flag=True,
    help=query_dict(manifest_commands, ("manifest", "get", "use_annotations")),
)
@click.option(
    "-oa",
    "--oauth",
    is_flag=True,
    help=query_dict(manifest_commands, ("manifest", "get", "oauth")),
)
@click.option(
    "-js",
    "--json_schema",
    help=query_dict(manifest_commands, ("manifest", "get", "json_schema")),
)
@click.option(
    "-av",
    "--alphabetize_valid_values",
    default = 'ascending',
    help=query_dict(manifest_commands, ("manifest", "get", "alphabetize_valid_values")),
)
@click.pass_obj
def get_manifest(
    ctx,
    title,
    data_type,
    jsonld,
    dataset_id,
    sheet_url,
    output_csv,
    use_annotations,
    oauth,
    json_schema,
    output_xlsx,
    alphabetize_valid_values,
):
    """
    Running CLI with manifest generation options.
    """
    # optional parameters that need to be passed to ManifestGenerator()
    # can be read from config.yml as well
    data_type = fill_in_from_config("data_type", data_type, ("manifest", "data_type"))
    jsonld = fill_in_from_config("jsonld", jsonld, ("model", "input", "location"))
    title = fill_in_from_config("title", title, ("manifest", "title"), allow_none=True)
    json_schema = fill_in_from_config(
        "json_schema",
        json_schema,
        ("model", "input", "validation_schema"),
        allow_none=True,
    )
    def create_single_manifest(data_type, output_csv=None, output_xlsx=None):
        # create object of type ManifestGenerator
        manifest_generator = ManifestGenerator(
            path_to_json_ld=jsonld,
            title=t,
            root=data_type,
            oauth=oauth,
            use_annotations=use_annotations,
            alphabetize_valid_values=alphabetize_valid_values,
        )

        # call get_manifest() on manifest_generator
        # if output_xlsx gets specified, output_format = "excel"
        if output_xlsx: 
            output_format = "excel"

            # if file name is in the path, and that file does not exist
            if not os.path.exists(output_xlsx):
                if ".xlsx" or ".xls" in output_xlsx:
                    path = Path(output_xlsx)
                    output_path = path.parent.absolute()
                else: 
                    logger.error(f"{output_xlsx} does not exists. Please try a valid file path")

            else: 
                output_path = output_xlsx
        else: 
            output_format = None
            output_path = None

        result = manifest_generator.get_manifest(
            dataset_id=dataset_id, sheet_url=sheet_url, json_schema=json_schema, output_format = output_format, output_path = output_path
        )

        if sheet_url:
            logger.info("Find the manifest template using this Google Sheet URL:")
            click.echo(result)
        if output_csv is None and output_xlsx is None: 
            prefix, _ = os.path.splitext(jsonld)
            prefix_root, prefix_ext = os.path.splitext(prefix)
            if prefix_ext == ".model":
                prefix = prefix_root
            output_csv = f"{prefix}.{data_type}.manifest.csv"

        elif output_xlsx:
            logger.info(
                f"Find the manifest template using this Excel file path: {output_xlsx}"
            )
            return result
        export_manifest_csv(file_path=output_csv, manifest=result)
        logger.info(
                f"Find the manifest template using this CSV file path: {output_csv}"
            )
        return result

    if type(data_type) is str:
        data_type = [data_type]

    if data_type[0] == 'all manifests':
        sg = SchemaGenerator(path_to_json_ld=jsonld)
        component_digraph = sg.se.get_digraph_by_edge_type('requiresComponent')
        components = component_digraph.nodes()
        for component in components:
            t = f'{title}.{component}.manifest'
            result = create_single_manifest(data_type = component)
    else:
        for dt in data_type:
            if len(data_type) > 1 and not output_xlsx:
                t = f'{title}.{dt}.manifest'
            elif output_xlsx: 
                if ".xlsx" or ".xls" in output_xlsx:
                    title_with_extension = os.path.basename(output_xlsx)
                    t = title_with_extension.split('.')[0]
            else:
                t = title
            result = create_single_manifest(data_type = dt, output_csv=output_csv, output_xlsx=output_xlsx)

    return result

@manifest.command(
    "migrate", short_help=query_dict(manifest_commands, ("manifest", "migrate", "short_help"))
)
@click_log.simple_verbosity_option(logger)
# define the optional arguments
@click.option(
    "-ps",
    "--project_scope",
    default=None,
    callback=parse_synIDs,
    help=query_dict(manifest_commands, ("manifest", "migrate", "project_scope")),
)
@click.option(
    "-ap",
    "--archive_project",
    default=None,
    help=query_dict(manifest_commands, ("manifest", "migrate", "archive_project")),
)
@click.option(
    "-p", "--jsonld", help=query_dict(manifest_commands, ("manifest", "get", "jsonld"))
)
@click.option(
    "-re",
    "--return_entities",
    is_flag=True,
    default=False,
    help=query_dict(manifest_commands, ("manifest", "migrate", "return_entities")),
)
@click.option(
    "-dr",
    "--dry_run",
    is_flag=True,
    default=False,
    help=query_dict(manifest_commands, ("manifest", "migrate", "dry_run")),
)
@click.pass_obj
def migrate_manifests(
    ctx,
    project_scope: List,
    archive_project: str,
    jsonld: str,
    return_entities: bool,
    dry_run: bool,
):
    """
    Running CLI with manifest migration options.
    """
    jsonld = fill_in_from_config("jsonld", jsonld, ("model", "input", "location"))

    access_token = os.getenv("SYNAPSE_ACCESS_TOKEN")
    full_scope = project_scope + [archive_project]
    if access_token:
        synStore = SynapseStorage(access_token = access_token, project_scope = full_scope)
    else:
        synStore = SynapseStorage(project_scope = full_scope)  

    for project in project_scope:
        if not return_entities:
            logging.info("Re-uploading manifests as tables")
            synStore.upload_annotated_project_manifests_to_synapse(project, jsonld, dry_run)
        if archive_project:
            logging.info("Migrating entitites")
            synStore.move_entities_to_new_project(project, archive_project, return_entities, dry_run)
        
    return 


import os
import string
import re
import io
import requests
import logging

from typing import (
    Any,
    Dict,
    Optional,
    Text,
)  # allows specifying explicit variable types

import pandas as pd
import numpy as np

from schematic.schemas.explorer import SchemaExplorer
from schematic import LOADER

from schematic.utils.validate_rules_utils import validate_schema_rules
from schematic.utils.df_utils import load_df

logger = logging.getLogger(__name__)


"""
Utility for converting csv file containing a data model definition schema (see scRNA-seq.csv for an example) into schema.org schema.
"""

# required headers for schema; may or may not abstract further; for now hardcode
required_headers = set(
    [
        "Attribute",
        "Description",
        "Valid Values",
        "DependsOn",
        "Required",
        "Parent",
        "Properties",
        "DependsOn Component",
        "Source",
        "Validation Rules",
    ]
)


def get_class(
    se: SchemaExplorer,
    class_display_name: str,
    description: str = None,
    subclass_of: list = None,
    requires_dependencies: list = None,
    requires_range: list = None,
    requires_components: list = None,
    required: bool = None,
    validation_rules: list = None,
) -> dict:

    """Constructs a new schema.org compliant class given a set of schema object attributes

    Args:
       se: a schema explorer object allowing the traversal and modification of a schema graph
       display_class_name: human readable label for the schema object/attribute: key characteristic X of the assay, related protocol, or downstream data that we want to record as metadata feature
       description: definition or a reference containing the definition of attribute X. Preferably provide a source ontology link or code in addition to the definition.
       subclass_of: *schema* label of this attribute/object's parent node in the schema
       requires_dependencies: important characteristics, if any, of attribute X that need to be recorded as metadata features given attribute X is specified. These characteristics are attributes themselves and need to pre-exist in the schema as such
       requires_range: a set/range of values that this attribute can be assigned to. this domain is stored in the rangeIncludes property of this object.
       requires_components: a set of associated components/categories that this object/entity requires for its full specification; each component is a high level ontology class in which entities/objects are categorized/componentized and it is an entity on its own that needs to exist in the schema.
       required: indicates if this attribute is required or optional in a schema
       validation_rules: a list of validation rules defined for this class (e.g. defining what is a valid object of this class)

    Returns: a json schema.org object
    """

    class_name = se.get_class_label_from_display_name(class_display_name)

    # setup biothings object template with mandatory elements
    class_attributes = {
        "@id": "bts:" + class_name,
        "@type": "rdfs:Class",
        "rdfs:comment": description
        if description and not pd.isnull(description)
        else "TBD",
        "rdfs:label": class_name,
        "schema:isPartOf": {"@id": "http://schema.biothings.io"},
    }

    # determine parent class of element and add subclass relationship to schema - required by biothings
    # if no subclass is provided, set a default to schema.org Thing
    if subclass_of:
        if len(subclass_of) == 1 and pd.isnull(subclass_of[0]):
            parent = {"rdfs:subClassOf": [{"@id": "schema:Thing"}]}
        else:
            parent = {
                "rdfs:subClassOf": [
                    {"@id": "bts:" + se.get_class_label_from_display_name(sub)}
                    for sub in subclass_of
                ]
            }
    else:
        parent = {"rdfs:subClassOf": [{"@id": "schema:Thing"}]}

    class_attributes.update(parent)

    # add optional attribute specifying attributes/objects that are required for the specification of this object
    # useful for specifying annotation requirements, for example
    if requires_dependencies:
        requirement = {
            "sms:requiresDependency": [
                {"@id": "bts:" + dep} for dep in requires_dependencies
            ]
        }
        class_attributes.update(requirement)

    # add optional attribute specifying the possible values this object can be set to; can be other objects, including primitives
    if requires_range:
        value_constraint = {
            "schema:rangeIncludes": [
                {"@id": "bts:" + se.get_class_label_from_display_name(val)}
                for val in requires_range
            ]
        }
        class_attributes.update(value_constraint)

    # add optional attribute specifying validation patterns associated with this object (e.g. precise definition of the object range)
    if validation_rules:
        class_attributes.update({"sms:validationRules": validation_rules})
    else:
        class_attributes.update({"sms:validationRules": []})

    # add optional attribute specifying the required components (i.e. high level ontology class in which entities/objects are categorized/componentized)
    # that are required for the specification of this object
    if requires_components:
        requirement = {
            "sms:requiresComponent": [{"@id": "bts:" + c} for c in requires_components]
        }
        class_attributes.update(requirement)

    if required:
        class_attributes.update({"sms:required": "sms:true"})
    else:
        class_attributes.update({"sms:required": "sms:false"})

    # ensure display name does not contain leading/trailing white spaces
    class_attributes.update({"sms:displayName": class_display_name.strip()})

    return class_attributes


def get_property(
    se: SchemaExplorer,
    property_display_name: str,
    property_class_name: str,
    description: str = None,
    requires_range: list = None,
    requires_dependencies: list = None,
    required: bool = None,
    validation_rules: str = None,
) -> dict:

    """Constructs a new schema.org compliant property of an existing schema.org object/class; note that the property itself is a schema.org object class.

    Args:
        se: a schema explorer object allowing the traversal and modification of a schema graph
        property_display_name: human readable label for the schema object/attribute: key characteristic X of the assay, related protocol, or downstream data that we want to record as metadata feature
        property_class_name: *schema* label of the class/object that this is a property of
        description: definition or a reference containing the definition of attribute X. Preferably provide a source ontology link or code in addition to the definition.
        requires_range: what is the set/domain of values that this attribute can be assigned to; currently only used to specify primitive types. TODO: extend to reg exp patterns
        requires_dependencies: important characteristics, if any, of property X that need to be recorded as metadata features given property X is specified. These characteristics are attributes themselves and need to pre-exist in the schema as such
        validation_rules: a list of validation rules defined for this class (e.g. defining what is a valid object of this property)


    Returns: a json schema.org  property object
    """
    property_name = se.get_property_label_from_display_name(property_display_name)

    property_attributes = {
        "@id": "bts:" + property_name,
        "@type": "rdf:Property",
        "rdfs:comment": description
        if description and not pd.isnull(description)
        else "TBD",
        "rdfs:label": property_name,
        "sms:displayName": property_display_name,
        "schema:domainIncludes": {
            "@id": "bts:" + se.get_class_label_from_display_name(property_class_name)
        },
        "schema:isPartOf": {"@id": "http://schema.biothings.io"},
    }
    if requires_range:
        value_constraint = {
            "schema:rangeIncludes": [
                {"@id": "bts:" + se.get_class_label_from_display_name(val)}
                for val in requires_range
            ]
        }
        property_attributes.update(value_constraint)

    if requires_dependencies:
        requirement = {
            "sms:requiresDependency": [
                {"@id": "bts:" + dep} for dep in requires_dependencies
            ]
        }
        property_attributes.update(requirement)

    # add optional attribute specifying validation patterns associated with this object (e.g. precise definition of the object range)
    if validation_rules:
        property_attributes.update({"sms:validationRules": validation_rules})
    else:
        property_attributes.update({"sms:validationRules": []})

    if required:
        property_attributes.update({"sms:required": "sms:true"})
    else:
        property_attributes.update({"sms:required": "sms:false"})

    #'http://schema.org/domainIncludes':{'@id': 'bts:' + property_class_name},
    #'http://schema.org/rangeIncludes':{'@id': 'schema:' + allowed_values},

    # ensure display name does not contain leading/trailing white spaces
    property_attributes.update({"sms:displayName": property_display_name.strip()})

    return property_attributes


def attribute_exists(se: SchemaExplorer, attribute_label: str) -> bool:

    """Check if a given attribute exists already in schema

    Args:
       se: a schema explorer object allowing the traversal and modification of a schema graph
       attribute_label: a schema label for the attribute to check

    Returns:
       True/False indicating if attribute exists or not
    """
    schema_graph = se.get_nx_schema()

    if attribute_label in schema_graph.nodes:
        return True
    return False


def check_schema_definition(schema_definition: pd.DataFrame) -> bool:

    """Checks if a schema definition data frame contains the right required headers.

    See schema definition guide for more details
    TODO: post and link schema definition guide

    Args:
        schema_definition: a pandas dataframe containing schema definition; see example here: https://docs.google.com/spreadsheets/d/1J2brhqO4kpeHIkNytzlqrdIiRanXDr6KD2hqjOTC9hs/edit#gid=0
    Raises: Exception
    """

    if required_headers.issubset(set(list(schema_definition.columns))):
        return
    elif "Requires" in list(schema_definition.columns) or "Requires Component" in list(
        schema_definition.columns
    ):
        raise ValueError(
            "The input CSV schema file contains the 'Requires' and/or the 'Requires "
            "Component' column headers. These columns were renamed to 'DependsOn' and "
            "'DependsOn Component', respectively. Switch to the new column names."
        )


def create_nx_schema_objects(
    schema_extension: pd.DataFrame, se: SchemaExplorer
) -> SchemaExplorer:
    """Creates classes for all attributes and adds them to the schema.
    Args:
        schema_extension: a pandas dataframe containing schema definition; see example here: https://docs.google.com/spreadsheets/d/1J2brhqO4kpeHIkNytzlqrdIiRanXDr6KD2hqjOTC9hs/edit#gid=0
        se: a schema explorer object allowing the traversal and modification of a schema graph
        base_schema_path: a path to a json-ld file containing an existing schema
    Returns:
        An updated schema explorer object
    """

    try:
        check_schema_definition(schema_extension)
        logger.debug("Schema definition csv ready for processing!")
    except:
        raise ValueError(
            f"Schema extension headers: {set(list(schema_extension.columns))} "
            f"do not match required schema headers: {required_headers}"
        )

    rel_dict = {
        "rdfs:subClassOf": {"parentOf": "in"},
        "schema:domainIncludes": {"domainValue": "in"},
        "sms:requiresDependency": {"requiresDependency": "out"},
        "sms:requiresComponent": {"requiresComponent": "out"},
        "schema:rangeIncludes": {"rangeValue": "out"},
    }

    # get attributes from Attribute column
    attributes = schema_extension[list(required_headers)].to_dict("records")

    # get all properties across all attributes from Properties column
    props = set(schema_extension[["Properties"]].dropna().values.flatten())

    # clean properties strings
    all_properties = []
    for prop in props:
        all_properties += [p.strip() for p in prop.split(",")]

    # get both attributes and their properties (if any)
    properties = schema_extension[["Attribute", "Properties"]].to_dict("records")

    # property to class map
    prop_2_class = {}
    for record in properties:
        if not pd.isnull(record["Properties"]):
            props = record["Properties"].strip().split(",")
            for p in props:
                prop_2_class[p.strip()] = record["Attribute"]

    logger.debug("Adding attributes")
    for attribute in attributes:

        required = None
        if not pd.isnull(attribute["Required"]):
            required = attribute["Required"]

        if not attribute["Attribute"] in all_properties:
            display_name = attribute["Attribute"]

            subclass_of = None
            if not pd.isnull(attribute["Parent"]):
                subclass_of = [
                    parent for parent in attribute["Parent"].strip().split(",")
                ]

            new_class = get_class(
                se,
                display_name,
                description=attribute["Description"],
                subclass_of=subclass_of,
                required=required,
            )

            se.add_schema_object_nx(new_class, **rel_dict)

            """
            print(se.get_nx_schema().nodes[new_class["rdfs:label"]])
            # check if attribute doesn't already exist and add it
            if not attribute_exists(se, new_class["rdfs:label"]):
                se.add_schema_object_nx(new_class, **rel_dict)
            else:
                print("ATTRIBUTE EXISTS")
                print(new_class)
            """

        else:
            display_name = attribute["Attribute"]

            new_property = get_property(
                se,
                display_name,
                prop_2_class[display_name],
                description=attribute["Description"],
                required=required,
            )
            # check if attribute doesn't already exist and add it
            if not attribute_exists(se, new_property["rdfs:label"]):
                se.add_schema_object_nx(new_property, **rel_dict)

    logger.debug("Done adding attributes")

    # TODO check if schema already contains property - may require property context in csv schema definition

    logger.debug("Adding and editing properties")

    for prop in properties:
        if not pd.isnull(prop["Properties"]):  # a class may have or not have properties
            for p in (
                prop["Properties"].strip().split(",")
            ):  # a class may have multiple properties
                attribute = prop["Attribute"]

                # check if property is already present as attribute under attributes column
                # TODO: adjust logic below to compactify code
                p = p.strip()
                if p in list(schema_extension["Attribute"]):
                    description = schema_extension.loc[
                        schema_extension["Attribute"] == p
                    ]["Description"].values[0]
                    property_info = se.explore_property(
                        se.get_property_label_from_display_name(p)
                    )
                    range_values = (
                        property_info["range"] if "range" in property_info else None
                    )
                    requires_dependencies = (
                        property_info["dependencies"]
                        if "dependencies" in property_info
                        else None
                    )
                    required = (
                        property_info["required"]
                        if "required" in property_info
                        else None
                    )

                    new_property = get_property(
                        se,
                        p,
                        property_info["domain"],
                        description=description,
                        requires_range=range_values,
                        requires_dependencies=requires_dependencies,
                        required=required,
                    )
                    se.edit_schema_object_nx(new_property)
                else:
                    description = None
                    new_property = get_property(
                        se, p, attribute, description=description
                    )
                    se.add_schema_object_nx(new_property, **rel_dict)

    logger.debug("Done adding properties")

    # # set range values and dependency requirements for each attribute
    # # if not already added, add each attribute in required values and dependencies to the schema extension
    # print("Editing attributes and properties to add requirements and value ranges")
    # print("====================================================================================")

    for attribute in attributes:

        # TODO: refactor processing of multi-valued cells in columns and corresponding schema updates; it would compactify code below if class and property are encapsulated as objects inheriting from a common attribute parent object

        # get values in range for this attribute, if any are specified
        range_values = attribute["Valid Values"]
        if not pd.isnull(range_values):
            # prepare the range values list and split based on appropriate delimiter
            # if the string "range_values" starts with double quotes, then extract all "valid values" within double quotes
            range_values_list = []
            if range_values[0] == '"':
                range_values_list = re.findall(r'"([^"]*)"', range_values)
            else:
                range_values_list = range_values.strip().split(",")

            for val in range_values_list:
                # check if value is in attributes column; add it as a class if not
                if not val.strip() in list(schema_extension["Attribute"]):

                    # determine parent class of the new value class
                    # if this attribute is not a property, set it as a parent class
                    if not attribute["Attribute"] in all_properties:
                        parent = attribute["Attribute"]
                    else:
                        # this attribute is a property, set the parent to the domain class of this attribute
                        parent = se.get_class_by_property(attribute["Attribute"])
                        if not parent:
                            raise ValueError(
                                f"Listed valid value: {val}, for attribute: {attribute['Attribute']} "
                                "must have a class parent. The extension could not be added to the schema."
                            )

                    new_class = get_class(
                        se, val, description=None, subclass_of=[parent]
                    )
                    # check if attribute doesn't already exist and add it
                    if not attribute_exists(se, new_class["rdfs:label"]):
                        se.add_schema_object_nx(new_class, **rel_dict)

                # update rangeIncludes of attribute
                # if attribute is not a property, then assume it is a class
                if not attribute["Attribute"] in all_properties:
                    class_info = se.explore_class(
                        se.get_class_label_from_display_name(attribute["Attribute"])
                    )
                    class_info["range"].append(
                        se.get_class_label_from_display_name(val)
                    )
                    class_range_edit = get_class(
                        se,
                        attribute["Attribute"],
                        description=attribute["Description"],
                        subclass_of=[attribute["Parent"]],
                        requires_dependencies=class_info["dependencies"],
                        requires_range=class_info["range"],
                        required=class_info["required"],
                        validation_rules=class_info["validation_rules"],
                    )
                    se.edit_schema_object_nx(class_range_edit)

                else:
                    # the attribute is a property
                    property_info = se.explore_property(
                        se.get_property_label_from_display_name(attribute["Attribute"])
                    )
                    property_info["range"].append(
                        se.get_class_label_from_display_name(val)
                    )
                    property_range_edit = get_property(
                        se,
                        attribute["Attribute"],
                        property_info["domain"],
                        description=property_info["description"],
                        requires_dependencies=property_info["dependencies"],
                        requires_range=property_info["range"],
                        required=property_info["required"],
                        validation_rules=property_info["validation_rules"],
                    )
                    se.edit_schema_object_nx(property_range_edit)

                logger.debug(val + " added to value range")

        # get validation rules for this attribute, if any are specified
        validation_rules = attribute["Validation Rules"]

        if not pd.isnull(validation_rules):
            
            # TODO: make validation rules delimiter configurable parameter
           
            validation_rules = [
                val_rule.strip() for val_rule in validation_rules.strip().split("::")
            ]

            validate_vr = validate_schema_rules(
                            validation_rules, 
                            attribute["Attribute"],
                            input_filetype = 'csv_schema')


            # update validation rules of attribute
            # if attribute is not a property, then assume it is a class
            if not attribute["Attribute"] in all_properties:
                class_info = se.explore_class(
                    se.get_class_label_from_display_name(attribute["Attribute"])
                )
                class_info["validation_rules"] = validation_rules
                class_val_rule_edit = get_class(
                    se,
                    attribute["Attribute"],
                    description=attribute["Description"],
                    subclass_of=[attribute["Parent"]],
                    requires_dependencies=class_info["dependencies"],
                    requires_range=class_info["range"],
                    required=class_info["required"],
                    validation_rules=class_info["validation_rules"],
                )
                se.edit_schema_object_nx(class_val_rule_edit)
            else:
                # the attribute is a property
                property_info = se.explore_property(
                    se.get_property_label_from_display_name(attribute["Attribute"])
                )
                property_info["validation_rules"] = validation_rules
                property_val_rule_edit = get_property(
                    se,
                    attribute["Attribute"],
                    property_info["domain"],
                    description=property_info["description"],
                    requires_dependencies=property_info["dependencies"],
                    requires_range=property_info["range"],
                    required=property_info["required"],
                    validation_rules=property_info["validation_rules"],
                )
                se.edit_schema_object_nx(property_val_rule_edit)
            try:
                logger.debug(val + "validation rules added")
            except:
                logger.debug("Validation rules added")

        # get dependencies for this attribute, if any are specified
        requires_dependencies = attribute["DependsOn"]
        if not pd.isnull(requires_dependencies):

            for dep in requires_dependencies.strip().split(","):
                # check if dependency is a property or not
                dep = dep.strip()
                dep_is_property = dep in all_properties
                dep_label = ""
                # set dependency label based on kind of dependency: class or property
                if dep_is_property:
                    dep_label = se.get_property_label_from_display_name(dep)
                else:
                    dep_label = se.get_class_label_from_display_name(dep)

                # check if dependency is in attributes column; add it to the list if not
                if not dep.strip() in list(schema_extension["Attribute"]):
                    # if dependency is a property create a new property; else create a new class
                    if not dep_is_property:
                        # if this attribute is not a property, set it as a parent class
                        if not attribute["Attribute"] in all_properties:
                            parent = attribute["Attribute"]
                        else:
                            # this attribute is a property, set the parent to the domain class of this attribute
                            parent = se.get_class_by_property(attribute["Attribute"])
                            if not parent:
                                raise ValueError(
                                    f"Listed required dependency: {dep}, for attribute: {attribute['Attribute']} "
                                    "must have a class parent. The extension could not be added to the schema."
                                )

                        new_class = get_class(
                            se, dep, description=None, subclass_of=[parent]
                        )
                        # se.add_schema_object_nx(new_class, **rel_dict)
                        # check if attribute doesn't already exist and add it
                        if not attribute_exists(se, new_class["rdfs:label"]):
                            se.add_schema_object_nx(new_class, **rel_dict)

                    else:
                        if not attribute["Attribute"] in all_properties:
                            domain_attribute = attribute["Attribute"]
                        else:
                            # this attribute is a property, set the domain of this property to the domain class of the attribute
                            domain_attribute = se.get_class_by_property(
                                attribute["Attribute"]
                            )
                            if not domain_attribute:
                                raise ValueError(
                                    f"Listed required dependency: {dep}, must have a class parent. "
                                    "The extension could not be added to the schema."
                                )

                        description = None
                        new_property = get_property(
                            se, dep, domain_attribute, description=description
                        )
                        # check if attribute doesn't already exist and add it
                        if not attribute_exists(se, new_property["rdfs:label"]):
                            se.add_schema_object_nx(new_property, **rel_dict)

                # update required dependencies of attribute
                # if attribute is not a property then assume it is a class
                if not attribute["Attribute"] in all_properties:
                    class_info = se.explore_class(
                        se.get_class_label_from_display_name(attribute["Attribute"])
                    )
                    class_info["dependencies"].append(dep_label)
                    class_dependencies_edit = get_class(
                        se,
                        attribute["Attribute"],
                        description=attribute["Description"],
                        subclass_of=[attribute["Parent"]],
                        requires_dependencies=class_info["dependencies"],
                        requires_range=class_info["range"],
                        required=class_info["required"],
                        validation_rules=class_info["validation_rules"],
                    )
                    se.edit_schema_object_nx(class_dependencies_edit)
                else:
                    # the attribute is a property then update as a property
                    property_info = se.explore_property(
                        se.get_property_label_from_display_name(attribute["Attribute"])
                    )
                    property_info["dependencies"].append(dep_label)
                    property_dependencies_edit = get_property(
                        se,
                        attribute["Attribute"],
                        property_info["domain"],
                        description=property_info["description"],
                        requires_dependencies=property_info["dependencies"],
                        requires_range=property_info["range"],
                        required=property_info["required"],
                        validation_rules=property_info["validation_rules"],
                    )
                    se.edit_schema_object_nx(property_dependencies_edit)

                logger.debug(dep + " added to dependencies")

            # TODO check for cycles in attribute dependencies schema subgraph

        # check if the attribute requires any components
        if not pd.isnull(attribute["DependsOn Component"]):
            component_dependencies = attribute["DependsOn Component"]
        else:
            continue

        # iterate over potentially multiple dependency components
        for comp_dep in component_dependencies.strip().split(","):

            # check if a component is already defined as an attribute; if not define it in the schema
            if not comp_dep.strip() in list(schema_extension["Attribute"]):

                # component is not in csv schema so try adding it as a class with a parent Thing
                new_class = get_class(se, comp_dep, description=None)

                # check if attribute doesn't already exist in schema.org schema and add it
                # (component may not be in csv schema, but could be in the base schema we are extending)
                if not attribute_exists(se, new_class["rdfs:label"]):
                    se.add_schema_object_nx(new_class, **rel_dict)

            # update this attribute requirements to include component
            class_info = se.explore_class(
                se.get_class_label_from_display_name(attribute["Attribute"])
            )
            class_info["component_dependencies"].append(
                se.get_class_label_from_display_name(comp_dep)
            )
            class_component_dependencies_edit = get_class(
                se,
                attribute["Attribute"],
                description=class_info["description"],
                subclass_of=class_info["subClassOf"],
                requires_dependencies=class_info["dependencies"],
                requires_range=class_info["range"],
                validation_rules=class_info["validation_rules"],
                requires_components=class_info["component_dependencies"],
            )
            se.edit_schema_object_nx(class_component_dependencies_edit)

        logger.debug(comp_dep + " added to dependencies")

        # TODO check for cycles in component dependencies schema subgraph

    logger.info("Done adding requirements and value ranges to attributes")

    return se


def _get_base_schema_path(base_schema: str = None) -> str:
    """Evaluate path to base schema.

    Args:
        base_schema: Path to base data model. BioThings data model is loaded by default.

    Returns:
        base_schema_path: Path to base schema based on provided argument.
    """
    biothings_schema_path = LOADER.filename("data_models/biothings.model.jsonld")
    base_schema_path = biothings_schema_path if base_schema is None else base_schema

    return base_schema_path


def _convert_csv_to_data_model(
    schema_csv: str, base_schema: str = None
) -> SchemaExplorer:
    """Convert provided CSV spec. in CSV format to data model in JSON-LD format.

    Args:
        schema_csv: Path to CSV file containing data to be translated to
                    JSON-LD data model. Can be path to local CSV or URL.

    Returns:
        base_se: SchemaExplorer object which has updated properties
                 (base_se.schema and base_se.schema_nx).
    """
    # create data model from provided RFC
    rfc_df = load_df(schema_csv, data_model=True)

    # instantiate schema explorer
    base_se = SchemaExplorer()

    # determine base schema path
    base_schema_path = _get_base_schema_path(base_schema)

    # load base schema (BioThings)
    base_se.load_schema(base_schema_path)

    # call parser code that converts a dataframe of the RFC
    # specs. into a JSON-LD data model
    base_se = create_nx_schema_objects(rfc_df, base_se)

    return base_se



from typing import Any, Sequence


class MissingConfigValueError(Exception):
    """Exception raised when configuration value not provided in config file.

    Args:
        config_keys: tuple of keys as present in config file.
        message: custom/pre-defined error message to be returned.

    Returns:
        message.
    """

    def __init__(self, config_keys: Sequence[Any], message: str = None) -> str:
        config_keys_str = " > ".join(config_keys)
        self.message = (
            "The configuration value corresponding to the argument "
            f"({config_keys_str}) doesn't exist. "
            "Please provide a value in the configuration file."
        )

        if message:
            self.message = message

        super().__init__(self.message)

    def __str__(self):
        return f"{self.message}"


class MissingConfigAndArgumentValueError(Exception):
    """Exception raised when configuration value not provided in config file.

    Args:
        arg_name: CLI argument name.
        config_keys: tuple of keys as present in config file.
        message: custom/pre-defined error message to be returned.

    Returns:
        message.
    """

    def __init__(
        self, arg_name: str, config_keys: Sequence[Any], message: str = None
    ) -> str:
        config_keys_str = " > ".join(config_keys)
        self.message = (
            f"The value corresponding to the CLI argument '--{arg_name}'"
            " doesn't exist. "
            "Please provide a value for either the CLI argument or "
            f"({config_keys_str}) in the configuration file."
        )

        if message:
            self.message = message

        super().__init__(self.message)

    def __str__(self):
        return f"{self.message}"


class AccessCredentialsError(Exception):
    """Exception raised when provided access credentials cannot be resolved.

    Args:
        project: Platform/project (e.g., synID of a project)
        message: custom/pre-defined error message to be returned.

    Returns:
        message.
    """

    def __init__(self, project: str, message: str = None) -> str:
        self.message = (
            f"Your access to '{project}'' could not be resolved. "
            "Please check your credentials and try again."
        )

        if message:
            self.message = message

        super().__init__(self.message)

    def __str__(self):
        return f"{self.message}"



import os
import string
import json
import logging

from typing import Any, Dict, Optional, Text, List

import inflection
import networkx as nx

from rdflib import Graph, Namespace, plugin, query
from networkx.algorithms.cycles import find_cycle
from networkx.readwrite import json_graph

from schematic.utils.curie_utils import (
    expand_curies_in_schema,
    uri2label,
    extract_name_from_uri_or_curie,
)
from schematic.utils.general import find_duplicates
from schematic.utils.io_utils import load_default, load_json, load_schemaorg
from schematic.utils.schema_utils import (
    load_schema_into_networkx,
    node_attrs_cleanup,
    class_to_node,
    relationship_edges,
)
from schematic.utils.general import dict2list, unlist
from schematic.utils.viz_utils import visualize
from schematic.utils.validate_utils import (
    validate_class_schema,
    validate_property_schema,
    validate_schema,
)
from schematic.schemas.curie import uri2curie, curie2uri

namespaces = dict(rdf=Namespace("http://www.w3.org/1999/02/22-rdf-syntax-ns#"))


logger = logging.getLogger(__name__)


class SchemaExplorer:
    """Class for exploring schema"""

    def __init__(self):
        self.load_default_schema()

    def load_schema(self, schema):
        """Load schema and convert it to networkx graph"""
        self.schema = load_json(schema)
        self.schema_nx = load_schema_into_networkx(self.schema)

    def export_schema(self, file_path):
        with open(file_path, "w",encoding="utf8") as f:
            json.dump(self.schema, f, sort_keys=True, indent=4, ensure_ascii=False)

    def load_default_schema(self):
        """Load default schema, either schema.org or biothings"""
        self.schema = load_default()
        self.schema_nx = load_schema_into_networkx(self.schema)

    def get_nx_schema(self):
        return self.schema_nx

    def get_edges_by_relationship(
        self, class_label: str, relationship: str
    ) -> List[str]:
        """Get a list of out-edges of a node where the edges match a specifc type of relationship.

        i.e., the edges connecting a node to its neighbors are of relationship type -- "parentOf" (set of edges to children / sub-class nodes).
        Note: possible edge relationships are -- parentOf, rangeValue, requiresDependency.

        Args:
            node: the node whose edges we need to look at.
            relationship: the type of link(s) that the above node and its immediate neighbors share.

        Returns:
            List of edges that are connected to the node.
        """
        edges = []

        mm_graph = self.get_nx_schema()

        for (u, v, key, c) in mm_graph.out_edges(node, data=True, keys=True):
            if key == relationship:
                edges.append((u, v))

        return edges

    def get_descendants_by_edge_type(
        self,
        source_node: str,
        relationship: str,
        connected: bool = True,
        ordered: bool = False,
    ) -> List[str]:
        """Get all nodes that are descendants of a given source node, based on a specific type of edge / relationship type.

        Args:
            source_node: The node whose descendants need to be retreived.
            relationship: Edge / link relationship type with possible values same as in above docs.
            connected: If True, we need to ensure that all descendant nodes are reachable from the source node, i.e., they are part of the same connected component.
                       If False, the descendants could be in multiple connected components.
                       Default value is True.
            ordered: If True, the list of descendants will be topologically ordered.
                     If False, the list has no particular order (depends on the order in which the descendats were traversed in the subgraph).

        Returns:
            List of nodes that are descendants from a particular node (sorted / unsorted)
        """
        mm_graph = self.get_nx_schema()

        # if mm_graph.has_node(source_node):
        # get all nodes that are reachable from a specified root /source node in the data model

        root_descendants = nx.descendants(mm_graph, source_node)
        # else:
        # print("The specified source node could not be found im the Networkx graph.")
        # return []

        subgraph_nodes = list(root_descendants)
        subgraph_nodes.append(source_node)
        descendants_subgraph = mm_graph.subgraph(subgraph_nodes)

        # prune the descendants subgraph so as to include only those edges that match the relationship type
        rel_edges = []
        for (u, v, key, c) in descendants_subgraph.edges(data=True, keys=True):
            if key == relationship:
                rel_edges.append((u, v))

        relationship_subgraph = nx.DiGraph()
        relationship_subgraph.add_edges_from(rel_edges)

        descendants = relationship_subgraph.nodes()

        if not descendants:
            # return empty list if there are no nodes that are reachable from the source node based on this relationship type
            return []

        if connected and ordered:
            # get the set of reachable nodes from the source node
            descendants = nx.descendants(relationship_subgraph, source_node)
            descendants.add(source_node)

            # normally, the descendants from a node are unordered (peculiarity of nx descendants call)
            # form the subgraph on descendants and order it topologically
            # this assumes an acyclic subgraph
            descendants = nx.topological_sort(
                relationship_subgraph.subgraph(descendants)
            )
        elif connected:
            # get the nodes that are reachable from a given source node
            # after the pruning process above some nodes in the root_descendants subgraph might have become disconnected and will be omitted
            descendants = nx.descendants(relationship_subgraph, source_node)
            descendants.add(source_node)
        elif ordered:
            # sort the nodes topologically
            # this requires the graph to be an acyclic graph
            descendants = nx.topological_sort(relationship_subgraph)

        return list(descendants)

    def get_adjacent_nodes_by_relationship(
        self, node: str, relationship: str
    ) -> List[str]:
        """Get a list of nodes that is / are adjacent to a given node, based on a relationship type.

        Args:
            node: the node whose edges we need to look at.
            relationship: the type of link(s) that the above node and its immediate neighbors share.

        Returns:
            List of nodes that are adjacent to the given node.
        """
        nodes = set()

        mm_graph = self.get_nx_schema()

        for (u, v, key, c) in mm_graph.out_edges(node, data=True, keys=True):
            if key == relationship:
                nodes.add(v)

        return list(nodes)

    def is_class_in_schema(self, class_label):
        if self.schema_nx.nodes[class_label]:
            return True
        else:
            return False

    def full_schema_graph(self, size=None):
        edges = self.schema_nx.edges()
        return visualize(edges, size=size)

    def sub_schema_graph(self, source, direction, size=None):
        if direction == "down":
            edges = list(nx.edge_bfs(self.schema_nx, [source]))
            return visualize(edges, size=size)
        elif direction == "up":
            paths = self.find_parent_classes(source)
            edges = []
            for _path in paths:
                _path.append(source)
                for i in range(0, len(_path) - 1):
                    edges.append((_path[i], _path[i + 1]))
            return visualize(edges, size=size)
        elif direction == "both":
            paths = self.find_parent_classes(source)
            edges = list(nx.edge_bfs(self.schema_nx, [source]))
            for _path in paths:
                _path.append(source)
                for i in range(0, len(_path) - 1):
                    edges.append((_path[i], _path[i + 1]))
            return visualize(edges, size=size)

    def find_parent_classes(self, schema_class):
        """Find all parents of the class"""

        digraph = self.get_digraph_by_edge_type("parentOf")

        root_node = list(nx.topological_sort(digraph))[0]
        # root_node = list(nx.topological_sort(self.schema_nx))[0]

        paths = nx.all_simple_paths(
            self.schema_nx, source=root_node, target=schema_class
        )
        # print(root_node)
        return [_path[:-1] for _path in paths]

    def find_class_specific_properties(self, schema_class):
        """Find properties specifically associated with a given class"""
        schema_uri = self.schema_nx.nodes[schema_class]["uri"]
        properties = []
        for record in self.schema["@graph"]:
            if record["@type"] == "rdf:Property":
                if (
                    type(record["schema:domainIncludes"]) == dict
                    and record["schema:domainIncludes"]["@id"] == schema_uri
                ):
                    properties.append(record["rdfs:label"])
                elif (
                    type(record["schema:domainIncludes"]) == list
                    and [
                        item
                        for item in record["schema:domainIncludes"]
                        if item["@id"] == schema_uri
                    ]
                    != []
                ):
                    properties.append(record["rdfs:label"])
        return properties

    def find_all_class_properties(self, schema_class, display_as_table=False):
        """Find all properties associated with a given class
        # TODO : need to deal with recursive paths
        """
        parents = self.find_parent_classes(schema_class)
        # print(schema_class)
        # print(parents)
        properties = [
            {
                "class": schema_class,
                "properties": self.find_class_specific_properties(schema_class),
            }
        ]
        for path in parents:
            path.reverse()
            for _parent in path:
                # print(_parent)
                properties.append(
                    {
                        "class": _parent,
                        "properties": self.find_class_specific_properties(_parent),
                    }
                )
        if not display_as_table:
            return properties
        else:
            content = [["Property", "Expected Type", "Description", "Class"]]
            for record in properties:
                for _property in record["properties"]:
                    property_info = self.explore_property(_property)
                    if "range" in property_info:
                        content.append(
                            [
                                _property,
                                property_info["range"],
                                property_info["description"],
                                record["class"],
                            ]
                        )
                    else:
                        content.append(
                            [_property, property_info["description"], record["class"]]
                        )

            # TODO: Log content

    def find_class_usages(self, schema_class):
        """Find where a given class is used as a value of a property"""
        usages = []
        schema_uri = self.schema_nx.nodes[schema_class]["uri"]
        for record in self.schema["@graph"]:
            usage = {}
            if record["@type"] == "rdf:Property":
                if "schema:rangeIncludes" in record:
                    p_range = dict2list(record["schema:rangeIncludes"])
                    for _doc in p_range:
                        if _doc["@id"] == schema_uri:
                            usage["property"] = record["rdfs:label"]
                            p_domain = dict2list(record["schema:domainIncludes"])
                            usage["property_used_on_class"] = unlist(
                                [self.uri2label(record["@id"]) for record in p_domain]
                            )
                            usage["description"] = record["rdfs:comment"]
            if usage:
                usages.append(usage)
        return usages

    def find_child_classes(self, schema_class):
        """Find schema classes that inherit from the given class"""
        return unlist(list(self.schema_nx.successors(schema_class)))

    def find_adjacent_child_classes(self, schema_class):

        return self.get_adjacent_nodes_by_relationship(schema_class, "parentOf")

    def explore_class(self, schema_class):
        """Find details about a specific schema class"""
        parents = []
        if "subClassOf" in self.schema_nx.nodes[schema_class]:
            schema_node_val = self.schema_nx.nodes[schema_class]["subClassOf"]

            parents_list = []
            if isinstance(schema_node_val, dict):
                parents_list.append(self.schema_nx.nodes[schema_class]["subClassOf"])
            else:
                parents_list = schema_node_val

            for parent in parents_list:
                parents.append(extract_name_from_uri_or_curie(parent["@id"]))

        requires_range = []
        if "rangeIncludes" in self.schema_nx.nodes[schema_class]:
            schema_node_val = self.schema_nx.nodes[schema_class]["rangeIncludes"]

            if isinstance(schema_node_val, dict):
                subclass_list = []
                subclass_list.append(
                    self.schema_nx.nodes[schema_class]["rangeIncludes"]
                )
            else:
                subclass_list = schema_node_val

            for range_class in subclass_list:
                requires_range.append(
                    extract_name_from_uri_or_curie(range_class["@id"])
                )

        requires_dependencies = []
        if "requiresDependency" in self.schema_nx.nodes[schema_class]:
            schema_node_val = self.schema_nx.nodes[schema_class]["requiresDependency"]

            if isinstance(schema_node_val, dict):
                subclass_list = []
                subclass_list.append(
                    self.schema_nx.nodes[schema_class]["requiresDependency"]
                )
            else:
                subclass_list = schema_node_val

            for dep_class in subclass_list:
                requires_dependencies.append(
                    extract_name_from_uri_or_curie(dep_class["@id"])
                )

        requires_components = []
        if "requiresComponent" in self.schema_nx.nodes[schema_class]:
            schema_node_val = self.schema_nx.nodes[schema_class]["requiresComponent"]

            if isinstance(schema_node_val, dict):
                subclass_list = []
                subclass_list.append(
                    self.schema_nx.nodes[schema_class]["requiresComponent"]
                )
            else:
                subclass_list = schema_node_val

            for comp_dep_class in subclass_list:
                requires_components.append(
                    extract_name_from_uri_or_curie(comp_dep_class["@id"])
                )

        required = False
        if "required" in self.schema_nx.nodes[schema_class]:
            required = self.schema_nx.nodes[schema_class]["required"]

        validation_rules = []
        if "validationRules" in self.schema_nx.nodes[schema_class]:
            validation_rules = self.schema_nx.nodes[schema_class]["validationRules"]

        # TODO: make class_info keys here the same as keys in schema graph nodes(e.g. schema_class above); note that downstream code using explore_class would have to be updated as well (e.g. csv_2_schemaorg)

        class_info = {
            "properties": self.find_class_specific_properties(schema_class),
            "description": self.schema_nx.nodes[schema_class]["description"],
            "uri": curie2uri(self.schema_nx.nodes[schema_class]["uri"], namespaces),
            #'usage': self.find_class_usages(schema_class),
            "usage": "NA",
            "child_classes": self.find_adjacent_child_classes(schema_class),
            "subClassOf": parents,
            "range": requires_range,
            "dependencies": requires_dependencies,
            "validation_rules": validation_rules,
            "required": required,
            "component_dependencies": requires_components,
            "parent_classes": parents
            #'parent_classes': self.find_parent_classes(schema_class)
        }

        if "displayName" in self.schema_nx.nodes[schema_class]:
            class_info["displayName"] = self.schema_nx.nodes[schema_class][
                "displayName"
            ]

        return class_info

    def get_class_validation_rules(self,class_label):
        rules=[]
        class_info = self.explore_class(class_label)
        
        if 'validation_rules' in class_info:
            rules=class_info['validation_rules']

        return rules

    def get_property_label_from_display_name(self, display_name, strict_camel_case = False):
        """Convert a given display name string into a proper property label string"""
        """
        label = ''.join(x.capitalize() or ' ' for x in display_name.split(' '))
        label = label[:1].lower() + label[1:] if label else ''
        """
        # This is the newer more strict method
        if strict_camel_case:
            display_name = display_name.strip().translate({ord(c): "_" for c in string.whitespace})
            label = inflection.camelize(display_name, uppercase_first_letter=False)

        # This method remains for backwards compatibility
        else:
            display_name = display_name.translate({ord(c): None for c in string.whitespace})
            label = inflection.camelize(display_name.strip(), uppercase_first_letter=False)

        return label

    def get_class_label_from_display_name(self, display_name, strict_camel_case = False):
        """Convert a given display name string into a proper class label string"""
        """
        label = ''.join(x.capitalize() or ' ' for x in display_name.split(' '))"""
        # This is the newer more strict method
        if strict_camel_case:
            display_name = display_name.strip().translate({ord(c): "_" for c in string.whitespace})
            label = inflection.camelize(display_name, uppercase_first_letter=True)

        # This method remains for backwards compatibility
        else:
            display_name = display_name.translate({ord(c): None for c in string.whitespace})
            label = inflection.camelize(display_name.strip(), uppercase_first_letter=True)

        return label

    def get_class_by_property(self, property_display_name):
        schema_property = self.get_property_label_from_display_name(
            property_display_name
        )

        for record in self.schema["@graph"]:
            if record["@type"] == "rdf:Property":
                if record["rdfs:label"] == schema_property:
                    p_domain = dict2list(record["schema:domainIncludes"])
                    return unlist(
                        [
                            self.uri2label(schema_class["@id"])
                            for schema_class in p_domain
                        ]
                    )

        return None

    def uri2label(self, uri):
        return uri.split(":")[1]

    def explore_property(self, schema_property):
        """Find details about a specific property
        TODO: refactor so that explore class and explore property reuse logic - they are *very* similar
        """
        property_info = {}
        for record in self.schema["@graph"]:
            if record["@type"] == "rdf:Property":
                if record["rdfs:label"] == schema_property:
                    property_info["id"] = record["rdfs:label"]
                    property_info["description"] = record["rdfs:comment"]
                    property_info["uri"] = curie2uri(record["@id"], namespaces)

                    p_domain = dict2list(record["schema:domainIncludes"])
                    property_info["domain"] = unlist(
                        [self.uri2label(record["@id"]) for record in p_domain]
                    )
                    if "schema:rangeIncludes" in record:
                        p_range = dict2list(record["schema:rangeIncludes"])
                        property_info["range"] = [
                            self.uri2label(record["@id"]) for record in p_range
                        ]
                    else:
                        property_info["range"] = []

                    if "sms:required" in record:
                        if "sms:true" == record["sms:required"]:
                            property_info["required"] = True
                        else:
                            property_info["required"] = False

                    validation_rules = []
                    if "sms:validationRules" in record:
                        property_info["validation_rules"] = record[
                            "sms:validationRules"
                        ]

                    if "sms:requiresDependency" in record:
                        p_dependencies = dict2list(record["sms:requiresDependency"])
                        property_info["dependencies"] = [
                            self.uri2label(record["@id"]) for record in p_dependencies
                        ]
                    else:
                        property_info["dependencies"] = []

                    if "sms:displayName" in record:
                        property_info["displayName"] = record["sms:displayName"]

                    break

        # check if properties are added multiple times

        return property_info

    def generate_class_template(self):
        """Generate a template for schema class"""
        template = {
            "@id": "uri or curie of the class",
            "@type": "rdfs:Class",
            "rdfs:comment": "description of the class",
            "rdfs:label": "class label, should match @id",
            "rdfs:subClassOf": {"@id": "parent class, could be list"},
            "schema:isPartOf": {"@id": "http://schema.biothings.io"},
        }
        return template

    def generate_property_template(self):
        """Generate a template for schema property"""
        template = {
            "@id": "url or curie of the property",
            "@type": "rdf:Property",
            "rdfs:comment": "description of the property",
            "rdfs:label": "carmel case, should match @id",
            "schema:domainIncludes": {
                "@id": "class which use it as a property, could be list"
            },
            "schema:isPartOf": {"@id": "http://schema.biothings.io"},
            "schema:rangeIncludes": {
                "@id": "relates a property to a class that constitutes (one of) the expected type(s) for values of the property"
            },
        }
        return template

    def edit_class(self, class_info):
        """Edit an existing class into schema"""
        for i, schema_class in enumerate(self.schema["@graph"]):
            if schema_class["rdfs:label"] == class_info["rdfs:label"]:
                validate_class_schema(class_info)  # why are we doing this in a loop?

                self.schema["@graph"][i] = class_info
                break

        # TODO: do we actually need to validate the entire schema if a class is just edited and the class passes validation?
        # validate_schema(self.schema)

        logger.info(f"Edited the class {class_info['rdfs:label']} successfully.")
        self.schema_nx = load_schema_into_networkx(self.schema)

    def update_class(self, class_info):
        """Add a new class into schema"""
        # print(class_info)
        validate_class_schema(class_info)
        self.schema["@graph"].append(class_info)
        validate_schema(self.schema)
        logger.info(f"Updated the class {class_info['rdfs:label']} successfully.")
        self.schema_nx = load_schema_into_networkx(self.schema)

    def edit_property(self, property_info):
        """Edit an existing property into schema"""
        for i, schema_property in enumerate(self.schema["@graph"]):
            if schema_property["rdfs:label"] == property_info["rdfs:label"]:
                validate_property_schema(property_info)
                self.schema["@graph"][i] = property_info

                # TODO: check if properties are added/edited multiple times (e.g. look at explore_property)
                break

        validate_schema(self.schema)
        logger.info(f"Edited the property {property_info['rdfs:label']} successfully.")
        self.schema_nx = load_schema_into_networkx(self.schema)

    def update_property(self, property_info):
        """Add a new property into schema"""
        validate_property_schema(property_info)
        self.schema["@graph"].append(property_info)
        validate_schema(self.schema)
        logger.info(f"Updated the property {property_info['rdfs:label']} successfully.")

    def get_nodes_descendants(self, graph, component):
        """
        Return a list of nodes reachable from source in graph 
        graph: networkx graph object
        component: any given node
        """
        all_descendants = list(nx.descendants(graph, component))

        return all_descendants

    def get_nodes_ancestors(self, graph, component):
        """
        Return a list of nodes reachable from source in graph 
        graph: networkx graph object
        component: any given node
        """
        all_ancestors = list(nx.ancestors(graph, component))

        return all_ancestors

    def get_digraph_by_edge_type(self, edge_type):

        multi_digraph = self.schema_nx

        digraph = nx.DiGraph()
        for (u, v, key, c) in multi_digraph.edges(data=True, keys=True):
            if key == edge_type:
                digraph.add_edge(u, v)

        # print(nx.find_cycle(digraph, orientation = "ignore"))

        return digraph

    # version of edit_class() method that directly acts on the networkx graph
    def edit_schema_object_nx(self, schema_object: dict) -> None:
        node_to_replace = class_to_node(class_to_convert=schema_object)

        # get the networkx graph associated with the SchemaExplorer object in its current state
        schema_graph_nx = self.get_nx_schema()

        # outer loop to loop over all the nodes in the graph constructed from master schema
        for node, data in schema_graph_nx.nodes(data=True):

            # innner loop to loop over the single node that is to be replaced/edited in the master graph
            for replace_node, replace_data in node_to_replace.nodes(data=True):

                # find the node to be replaced in the graph
                if node == replace_node:

                    # for the "comment", "required", "displayName", "validationRules" fields/keys it's okay to do a direct replacement
                    # without having to worry about adding/removing any associated edges

                    # ques. is it more expensive to do a checking operation (diff b/w fields) or a replace operation?

                    if (
                        "comment" in data and "comment" in replace_data
                    ):  # replace contents of "comment" from replacement node
                        schema_graph_nx.nodes[node]["comment"] = node_to_replace.nodes[
                            replace_node
                        ]["comment"]
                        schema_graph_nx.nodes[node][
                            "description"
                        ] = node_to_replace.nodes[replace_node]["description"]

                    if (
                        "required" in data and "required" in replace_data
                    ):  # replace boolean value of "required" from replacement node
                        schema_graph_nx.nodes[node]["required"] = node_to_replace.nodes[
                            replace_node
                        ]["required"]

                    if (
                        "displayName" in data and "displayName" in replace_data
                    ):  # replace contents of "displayName" from replacement node
                        schema_graph_nx.nodes[node][
                            "displayName"
                        ] = node_to_replace.nodes[replace_node]["displayName"]

                    if (
                        "validationRules" in data and "validationRules" in replace_data
                    ):  # replace contents of "validationRules" from replacement node
                        schema_graph_nx.nodes[node][
                            "validationRules"
                        ] = node_to_replace.nodes[replace_node]["validationRules"]

                    # for the "subClassOf", "requiresDependency", "requiresComponent", "rangeIncludes" fields/keys require rejiggering
                    # of associated edges
                    # general strategy we follow for rejiggering is remove edges that existed formerly and add new edges based on contents
                    # of the replacement node

                    # "subClassOf" key related edge manipulation
                    if "subClassOf" in replace_data:

                        # if the "subClassOf" attribute already exists on the node, then remove all the "parentOf" in-edges
                        # associated with that node
                        if "subClassOf" in data:
                            # remove formerly existent edges from the master schema/graph
                            for (u, v) in list(schema_graph_nx.in_edges([node])):

                                # there are certain nodes which have "subClassOf" data in list format
                                if type(data["subClassOf"]) == list:
                                    for _edges_to_replace in data["subClassOf"]:
                                        edge_repl = extract_name_from_uri_or_curie(
                                            _edges_to_replace["@id"]
                                        )

                                        if edge_repl == u:

                                            try:
                                                # we need to make sure to remove only edges that are tagged with the "parentOf" label
                                                schema_graph_nx.remove_edges_from(
                                                    [(u, v, "parentOf")]
                                                )
                                            except TypeError:
                                                pass

                                # there are certain nodes which have "subClassOf" data in dict format
                                elif type(data["subClassOf"]) == dict:
                                    for k_id, v_curie in data["subClassOf"].items():
                                        edge_repl = extract_name_from_uri_or_curie(
                                            v_curie
                                        )

                                        if edge_repl == u:

                                            try:
                                                schema_graph_nx.remove_edges_from(
                                                    [(u, v, "parentOf")]
                                                )
                                            except TypeError:
                                                pass

                        # extract node names from replacement node and use it to add edges to the master schema/graph
                        parents = replace_data["subClassOf"]
                        if type(parents) == list:
                            for _parent in parents:
                                target_node = extract_name_from_uri_or_curie(
                                    _parent["@id"]
                                )

                                # label to be associated with "subClassOf" keys is "parentOf"
                                if target_node != replace_node:

                                    # make note of the fact that we are changing in-edges here
                                    schema_graph_nx.add_edge(
                                        target_node, replace_node, key="parentOf"
                                    )
                        elif type(parents) == dict:
                            for _k_parent, _v_parent in parents.items():
                                target_node = extract_name_from_uri_or_curie(_v_parent)

                                # label to be associated with "subClassOf" keys is "parentOf"
                                if target_node != replace_node:

                                    # make note of the fact that we are changing in-edges here
                                    schema_graph_nx.add_edge(
                                        target_node, replace_node, key="parentOf"
                                    )

                        # once the edges have been added, change the contents of the node
                        schema_graph_nx.nodes[node][
                            "subClassOf"
                        ] = node_to_replace.nodes[replace_node]["subClassOf"]

                    # "requiresDependency" key related edge manipulation
                    if "requiresDependency" in replace_data:

                        # if the "requiresDependency" attribute already exists on the node, then remove all the "requiresDependency" in-edges
                        # associated with that node
                        if "requiresDependency" in data:

                            for (u, v) in list(schema_graph_nx.out_edges([node])):
                                # there are certain nodes which have "requiresDependency" data in list format
                                if type(data["requiresDependency"]) == list:
                                    for _edges_to_replace in data["requiresDependency"]:
                                        edge_repl = extract_name_from_uri_or_curie(
                                            _edges_to_replace["@id"]
                                        )

                                        if edge_repl == v:

                                            try:
                                                schema_graph_nx.remove_edges_from(
                                                    [u, v, "requiresDependency"]
                                                )
                                            except TypeError:
                                                pass

                                # there are certain nodes which have "requiresDependency" data in dict format
                                elif type(data["requiresDependency"]) == dict:
                                    for k_id, v_curie in data[
                                        "requiresDependency"
                                    ].items():
                                        edge_repl = extract_name_from_uri_or_curie(
                                            v_curie
                                        )

                                        if edge_repl == u:

                                            try:
                                                schema_graph_nx.remove_edges_from(
                                                    [u, v, "requiresDependency"]
                                                )
                                            except TypeError:
                                                pass

                            deps = replace_data["requiresDependency"]
                            if type(deps) == list:
                                for _dep in deps:
                                    target_node = extract_name_from_uri_or_curie(
                                        _dep["@id"]
                                    )

                                    if target_node != replace_node:

                                        # make not of the fact that edges being added here are out-edges
                                        schema_graph_nx.add_edge(
                                            replace_node,
                                            target_node,
                                            key="requiresDependency",
                                        )
                            elif type(deps) == dict:
                                for _k_dep, _v_dep in deps.items():
                                    target_node = extract_name_from_uri_or_curie(_v_dep)

                                    if target_node != replace_node:

                                        # make not of the fact that edges being added here are out-edges
                                        schema_graph_nx.add_edge(
                                            replace_node,
                                            target_node,
                                            key="requiresDependency",
                                        )

                        schema_graph_nx.nodes[node][
                            "requiresDependency"
                        ] = node_to_replace.nodes[replace_node]["requiresDependency"]

                    # "requiresComponent" key related edge manipulation
                    if "requiresComponent" in replace_data:

                        if "requiresComponent" in data:
                            for (u, v) in list(schema_graph_nx.out_edges([node])):
                                # there are certain nodes which have "requiresComponent" data in list format
                                if type(data["requiresComponent"]) == list:
                                    for _edges_to_replace in data["requiresComponent"]:
                                        edge_repl = extract_name_from_uri_or_curie(
                                            _edges_to_replace["@id"]
                                        )

                                        if edge_repl == v:

                                            try:
                                                schema_graph_nx.remove_edges_from(
                                                    [u, v, "requiresComponent"]
                                                )
                                            except TypeError:
                                                pass

                                elif type(data["requiresComponent"]) == dict:
                                    for k_id, v_curie in data[
                                        "requiresComponent"
                                    ].items():
                                        edge_repl = extract_name_from_uri_or_curie(
                                            v_curie
                                        )

                                        if edge_repl == v:

                                            try:
                                                schema_graph_nx.remove_edges_from(
                                                    [u, v, "requiresComponent"]
                                                )
                                            except TypeError:
                                                pass

                        comps = replace_data["requiresComponent"]
                        if type(comps) == list:
                            for _comp in comps:
                                target_node = extract_name_from_uri_or_curie(
                                    _comp["@id"]
                                )

                                if target_node != replace_node:
                                    schema_graph_nx.add_edge(
                                        replace_node,
                                        target_node,
                                        key="requiresComponent",
                                    )
                        elif type(comps) == dict:
                            for _k_comp, _v_comp in deps.items():
                                target_node = extract_name_from_uri_or_curie(_v_comp)

                                if target_node != replace_node:

                                    # make not of the fact that edges being added here are out-edges
                                    schema_graph_nx.add_edge(
                                        replace_node,
                                        target_node,
                                        key="requiresDependency",
                                    )

                        schema_graph_nx.nodes[node][
                            "requiresComponent"
                        ] = node_to_replace.nodes[replace_node]["requiresComponent"]

                    # "rangeIncludes" key related edge manipulation
                    if "rangeIncludes" in replace_data:

                        if "rangeIncludes" in data:
                            for (u, v) in list(schema_graph_nx.out_edges([node])):
                                # there are certain nodes which have "rangeIncludes" data in list format
                                if type(data["rangeIncludes"]) == list:
                                    for _edges_to_replace in data["rangeIncludes"]:
                                        edge_repl = extract_name_from_uri_or_curie(
                                            _edges_to_replace["@id"]
                                        )

                                        if edge_repl == v:
                                            try:
                                                schema_graph_nx.remove_edges_from(
                                                    [u, v, "rangeIncludes"]
                                                )
                                            except TypeError:
                                                pass

                                elif type(data["rangeIncludes"]) == dict:
                                    for k_id, v_curie in data["rangeIncludes"].items():
                                        edge_repl = extract_name_from_uri_or_curie(
                                            v_curie
                                        )

                                        if edge_repl == v:
                                            try:
                                                schema_graph_nx.remove_edges_from(
                                                    [u, v, "rangeIncludes"]
                                                )
                                            except TypeError:
                                                pass

                        range_inc = replace_data["rangeIncludes"]
                        if type(range_inc) == list:
                            for _rinc in range_inc:
                                target_node = extract_name_from_uri_or_curie(
                                    _rinc["@id"]
                                )

                                if target_node != replace_node:
                                    schema_graph_nx.add_edge(
                                        replace_node, target_node, key="rangeValue"
                                    )
                        elif type(range_inc) == dict:
                            for _k_rinc, _v_rinc in deps.items():
                                target_node = extract_name_from_uri_or_curie(_v_rinc)

                                if target_node != replace_node:

                                    # make not of the fact that edges being added here are out-edges
                                    schema_graph_nx.add_edge(
                                        replace_node, target_node, key="rangeValue"
                                    )

                        schema_graph_nx.nodes[node][
                            "rangeIncludes"
                        ] = node_to_replace.nodes[replace_node]["rangeIncludes"]

        # set the networkx schema graph to the the modified networkx schema
        self.schema_nx = schema_graph_nx

        # print("Added node {} to the graph successfully.".format(schema_object["rdfs:label"]))

        # part of the code that replaces the modified class in the original JSON-LD schema (not in the data/ folder though)
        for i, schema_class in enumerate(self.schema["@graph"]):
            if schema_class["rdfs:label"] == schema_object["rdfs:label"]:
                # validate_class_schema(schema_object)    # validate that the class to be modified follows the structure for any generic class (node)

                self.schema["@graph"][i] = schema_object
                break

    # version of update_class() method that directly acts on the networkx graph
    def add_schema_object_nx(self, schema_object: dict, **kwargs: dict) -> None:
        node = node_attrs_cleanup(schema_object)

        if "required" in node:
            if "sms:true" == schema_object["sms:required"]:
                node["required"] = True
            else:
                node["required"] = False

        if "sms:validationRules" in schema_object:
            node["validationRules"] = schema_object["sms:validationRules"]
        else:
            node["validationRules"] = []

        node["uri"] = schema_object["@id"]
        node["description"] = schema_object["rdfs:comment"]

        # get the networkx graph associated with the SchemaExplorer object in its current state
        schema_graph_nx = self.get_nx_schema()

        # add node to graph
        schema_graph_nx.add_node(schema_object["rdfs:label"], **node)

        schema_graph_nx = relationship_edges(schema_graph_nx, schema_object, **kwargs)

        # set the networkx schema graph to the the modified networkx schema
        self.schema_nx = schema_graph_nx

        # print("Edited node {} successfully.".format(schema_object["rdfs:label"]))

        # update the JSON-LD schema after modifying the networkx graph
        # validate_class_schema(schema_object)
        self.schema["@graph"].append(schema_object)
        # validate_schema(self.schema)



import sys
import logging

import click
import click_log

from schematic.configuration import CONFIG
from schematic.loader import LOADER
from schematic.utils.google_api_utils import download_creds_file, generate_token
from schematic.utils.cli_utils import query_dict
from schematic.help import init_command

logging.basicConfig(
    format=("%(levelname)s: [%(asctime)s] %(name)s" " - %(message)s"),
    level=logging.WARNING,
    datefmt="%Y-%m-%d %H:%M:%S",
)


# Suppress INFO-level logging from some dependencies
logging.getLogger("keyring").setLevel(logging.ERROR)
logging.getLogger("rdflib").setLevel(logging.ERROR)

logger = logging.getLogger(__name__)
click_log.basic_config(logger)


@click.command("init", short_help=query_dict(init_command, ("init", "short_help")))
@click_log.simple_verbosity_option(logger)
@click.option(
    "-a",
    "--auth",
    default="token",
    type=click.Choice(["token", "service_account"], case_sensitive=False),
    help=query_dict(init_command, ("init", "auth")),
)
@click.option(
    "-c", "--config", required=True, help=query_dict(init_command, ("init", "config"))
)
def init(auth, config):
    """Initialize mode of authentication for schematic."""
    try:
        logger.debug(f"Loading config file contents in '{config}'")
        obj = CONFIG.load_config(config)
    except ValueError as e:
        logger.error("'--config' not provided or environment variable not set.")
        logger.exception(e)
        sys.exit(1)

    # download crdentials file based on selected mode of authentication
    download_creds_file(auth)

    # if authentication method is token-based
    # then create 'token.pickle' file from 'credentials.json' file
    if auth == "token":
        generate_token()
